import binascii
import logging
import typing

from paradox.config import config as cfg
from paradox.lib.crypto import encrypt, decrypt
from paradox.parsers.paradox_ip_messages import ip_message
from .connection import ConnectionProtocol
from ..lib.utils import call_soon_in_main_loop

logger = logging.getLogger('PAI').getChild(__name__)


def checksum(data, min_message_length):
    """Calculates the 8bit checksum of Paradox messages"""
    c = 0

    if data is None or len(data) < min_message_length:
        return False

    for i in data[:-1]:
        c += i

    r = (c % 256) == data[-1]
    return r


class SerialConnectionProtocol(ConnectionProtocol):
    def __init__(self, on_message: typing.Callable[[bytes], None], on_port_open, on_con_lost):
        super(SerialConnectionProtocol, self).__init__(on_message=on_message, on_con_lost=on_con_lost)
        self.buffer = b''
        self.on_port_open = on_port_open

    def connection_made(self, transport):
        super(SerialConnectionProtocol, self).connection_made(transport)
        self.on_port_open()

    async def _send_message(self, message):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> SER {}".format(binascii.hexlify(message)))

        await self.transport.write(message)

    def send_message(self, message):
        call_soon_in_main_loop(self._send_message(message))

    def data_received(self, recv_data):
        self.buffer += recv_data

        min_length = 4 if self.use_variable_message_length else 37

        while len(self.buffer) >= min_length:
            if self.use_variable_message_length:
                if self.buffer[0] >> 4 == 0:
                    potential_packet_length = 37
                elif self.buffer[0] >> 4 in [1, 3, 4, 5, 6, 7, 8, 9]:
                    potential_packet_length = self.buffer[1] if 0 < self.buffer[1] <= 71 else 37
                elif self.buffer[0] >> 4 in [0x0A, 0x0B, 0x0D]:
                    potential_packet_length = self.buffer[1]
                elif self.buffer[0] >> 4 == 0x0C:
                    potential_packet_length = self.buffer[1] * 256 + self.buffer[2]
                elif self.buffer[0] >> 4 == 0x0E:
                    if self.buffer[1] < 37 or self.buffer[1] == 0xFF:
                        # MG/SP in 21st century and EVO Live Events. Probable values=0x13, 0x13, 0x00, 0xFF
                        potential_packet_length = 37
                    else:
                        potential_packet_length = self.buffer[1]
                else:
                    potential_packet_length = 37

            else:
                potential_packet_length = 37

            if len(self.buffer) < potential_packet_length:
                break

            frame = self.buffer[:potential_packet_length]

            if checksum(frame, min_length):
                self.buffer = self.buffer[len(frame):]  # Remove message
                if cfg.LOGGING_DUMP_PACKETS:
                    logger.debug("SER -> PAI {}".format(binascii.hexlify(frame)))

                self.on_message(frame)
            else:
                self.buffer = self.buffer[1:]

    def connection_lost(self, exc):
        logger.error('The serial port was closed')
        self.buffer = b''
        super(SerialConnectionProtocol, self).connection_lost(exc)


class IPConnectionProtocol(ConnectionProtocol):
    def __init__(self, on_message: typing.Callable[[bytes], None], on_con_lost, key):
        super(IPConnectionProtocol, self).__init__(on_message=on_message, on_con_lost=on_con_lost)
        self.buffer = b''
        self.key = key

    def send_raw(self, raw):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> Mod {}".format(binascii.hexlify(raw)))
        self.transport.write(raw)

    def send_message(self, message):
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("PAI -> IPC {}".format(binascii.hexlify(message)))

        payload = encrypt(message, self.key)
        msg = ip_message.build(
            dict(header=dict(length=len(message), unknown0=0x04, flags=0x09, command=0x00, encrypt=1), payload=payload))
        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("IPC -> Mod {}".format(binascii.hexlify(msg)))
        self.transport.write(msg)

    def _get_message_payload(self, data):
        message = ip_message.parse(data)

        if len(message.payload) >= 16 and len(message.payload) % 16 == 0 and message.header.flags & 0x01 != 0:
            message_payload = decrypt(data[16:], self.key)[:message.header.length]
        else:
            message_payload = message.payload[:message.header.length]

        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("IPC -> PAI {}".format(binascii.hexlify(message_payload)))

        return message_payload

    def data_received(self, recv_data):
        self.buffer += recv_data

        if self.buffer[0] != 0xaa:
            if len(self.buffer) > 0:
                logger.warning('Dangling data in the receive buffer: %s' % binascii.hexlify(self.buffer))
            self.buffer = b''
            return

        if len(recv_data) + 16 < self.buffer[1]:
            return

        if len(self.buffer) % 16 != 0:
            return

        if cfg.LOGGING_DUMP_PACKETS:
            logger.debug("Mod -> IPC {}".format(binascii.hexlify(self.buffer)))

        self.on_message(self._get_message_payload(self.buffer))
        self.buffer = b''



import asyncio
import logging
from logging.handlers import RotatingFileHandler
import sys
import time
import signal

from paradox import VERSION
from paradox.config import config as cfg
from paradox.exceptions import AuthenticationFailed

from paradox.paradox import Paradox
from paradox.interfaces.interface_manager import InterfaceManager

alarm = None
interface_manager = None

logger = logging.getLogger('PAI')


def get_format(level):
    if level <= logging.DEBUG:
        return '%(asctime)s - %(levelname)-8s - %(threadName)-10s - %(name)s - %(message)s'
    else:
        return '%(asctime)s - %(levelname)-8s - %(name)s - %(message)s'


def config_logger(logger):
    logger_level = cfg.LOGGING_LEVEL_CONSOLE

    if cfg.LOGGING_FILE is not None:
        logfile_handler = RotatingFileHandler(
            cfg.LOGGING_FILE, mode='a',
            maxBytes=cfg.LOGGING_FILE_MAX_SIZE * 1024 * 1024,
            backupCount=cfg.LOGGING_FILE_MAX_FILES,
            encoding=None, delay=0
        )

        logfile_handler.setLevel(cfg.LOGGING_LEVEL_FILE)
        logfile_handler.setFormatter(logging.Formatter(get_format(logger_level)))
        logger.addHandler(logfile_handler)
        logger_level = min(logger_level, cfg.LOGGING_LEVEL_FILE)

    logconsole_handler = logging.StreamHandler()
    logconsole_handler.setLevel(cfg.LOGGING_LEVEL_CONSOLE)
    logconsole_handler.setFormatter(logging.Formatter(get_format(logger_level)))
    logger.addHandler(logconsole_handler)

    logger.setLevel(logger_level)


def exit_handler(signum=None, frame=None):
    global alarm, interface_manager

    if signum is not None:
        logger.info('Captured signal %d. Exiting' % signum)
    
    if alarm:
        alarm.disconnect()
        alarm = None

    if interface_manager:
        interface_manager.stop()
        interface_manager = None

    time.sleep(1)
    
    logger.info("Good bye!")
    sys.exit(0)


async def run_loop(alarm: Paradox):
    retry = 1
    while alarm is not None:
        logger.info("Starting...")
        retry_time_wait = 2 ^ retry
        retry_time_wait = 30 if retry_time_wait > 30 else retry_time_wait

        try:
            if await alarm.connect():
                retry = 1
                await alarm.loop()
            else:
                logger.error("Unable to connect to alarm")

            time.sleep(retry_time_wait)
        except ConnectionError as e:  # Connection to IP Module or MQTT lost
            logger.error("Connection to panel lost: %s. Restarting" % str(e))
            time.sleep(retry_time_wait)

        except OSError:  # Connection to IP Module or MQTT lost
            logger.exception("Restarting")
            time.sleep(retry_time_wait)

        except (KeyboardInterrupt, SystemExit, AuthenticationFailed):
            break  # break exits the retry loop
        except Exception:
            logger.exception("Restarting")
            time.sleep(retry_time_wait)

        retry += 1


def main(args):
    global alarm, interface_manager
    
    if 'config' in args and args.config is not None:
        import os
        config_file = os.path.abspath(args.config)
        cfg.load(config_file)
    else:
        cfg.load()

    config_logger(logger)

    logger.info(f"Starting Paradox Alarm Interface {VERSION}")
    logger.info(f"Config loaded from {cfg.CONFIG_FILE_LOCATION}")

    logger.info(f"Console Log level set to {cfg.LOGGING_LEVEL_CONSOLE}")

    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)

    # Start interacting with the alarm
    alarm = Paradox()
    interface_manager = InterfaceManager(alarm, config=cfg)
    interface_manager.start()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_loop(alarm))
    
    exit_handler()




from paradox.event import EventLevel, LiveEvent


def _toggle_provider(old):
    return not old


def _minor2_provider(event, *_):
    return event.minor2


def _request_status_refresh(alarm, *_, **__):
    alarm.request_status_refresh()


def _zone_generated_alarm(event, storage, *_, **__):
    assert isinstance(event, LiveEvent)
    storage.update_container_object('partition', event.partition, {
        "current_state": "triggered",
        "alarm_in_memory": True,
        "was_in_alarm": True,
        "audible_alarm": True,
        "exit_delay": False
    })


event_map = {
    0: dict(
        level=EventLevel.DEBUG, change=dict(open=False), type='zone', message='Zone {label} OK'),
    1: dict(
        level=EventLevel.DEBUG, change=dict(open=True), type='zone', message='Zone {label} open'),
    2: dict(
        level=EventLevel.WARN, tags=['trouble'], change=dict(zone_tamper_trouble=True), type='zone',
        message='Zone {label} tampered'),
    3: dict(
        level=EventLevel.WARN, tags=['trouble'], type='zone', change=dict(fire_loop_trouble=True),
        message='Zone {label} in fire loop trouble'),
    4: dict(
        level=EventLevel.DEBUG, message='Non-reportable event',
        sub={
            0: dict(level=EventLevel.CRITICAL, type='system', tags=['trouble'], message='Telephone line trouble'),
            1: dict(type='system', message='Smoke detector reset'),
            2: dict(level=EventLevel.INFO, tags=['arm'], type='partition', change=dict(arm=True),
                    message='Arm with no entry delay'),
            3: dict(level=EventLevel.INFO, tags=['arm'], type='partition', change=dict(arm_stay=True, arm=True),
                    message='Arm in stay mode'),
            4: dict(level=EventLevel.INFO, tags=['arm'], type='partition', change=dict(arm_away=True, arm=True),
                    message='Arm in Away mode'),
            5: dict(level=EventLevel.INFO, tags=['arm'], type='partition', change=dict(arm_stay=True, arm=True),
                    message='Full arm when in Stay mode'),
            6: dict(type='system', message='Voice module access'),
            7: dict(type='system', message='Remote control access by {@user}', user=_minor2_provider),
            8: dict(type='system', message='PC fail to communicate'),
            9: dict(type='system', message='Midnight'),
            10: dict(type='system', message='Neware user login'),
            11: dict(type='system', message='Neware user logout'),
            12: dict(type='system', message='User initiated call-up'),
            13: dict(type='system', message='Force answer'),
            14: dict(type='system', message='Force hangup'),
            15: dict(type='system', message='Reset to default'),
            16: dict(type='system', message='Auxiliary output manually activated'),
            17: dict(type='system', message='Auxiliary output manually deactivated'),
            18: dict(type='system', message='Voice reporting failed'),
            19: dict(type='system', message='Fail to communicate restore', telephone=_minor2_provider),
            20: dict(type='system', message='Software access (VDMP3, Ip100, Neware, WinLoad)'),  # minor2 has sub events
            21: dict(type='system', message='IPR512 1 registration status'),
            22: dict(type='system', message='IPR512 2 registration status'),
            23: dict(type='system', message='IPR512 3 registration status'),
            24: dict(type='system', message='IPR512 4 registration status')
        }),
    5: dict(
        level=EventLevel.INFO, type='user', message='User {label} code entered on keypad', keypad=_minor2_provider),
    6: dict(
        level=EventLevel.INFO, type='door', message='User/card access on door {label}', user=_minor2_provider),
    7: dict(
        level=EventLevel.INFO, type='user', message='Bypass programming access'),
    8: dict(
        level=EventLevel.WARN, type='zone', message='TX delay zone {label} alarm'),
    9: dict(
        level=EventLevel.INFO, tags=['arm'], change=dict(arm=True), type='partition',
        message='Arming {label} with master code'),
    10: dict(
        level=EventLevel.INFO, tags=['arm'], change=dict(arm=True), type='partition',
        message='Arming {label} with user code'),
    11: dict(
        level=EventLevel.INFO, tags=['arm'], change=dict(arm=True), type='partition',
        message='Arming {label} with keyswitch'),
    12: dict(
        level=EventLevel.INFO, tags=['arm'], change=dict(arm=True), type='partition', message='Special arming {label}',
        sub={  # no label passed
            0: dict(type='partition', message='auto arming'),
            1: dict(type='partition', message='arming with Winload by {@user}', user=_minor2_provider),
            2: dict(type='partition', message='late to close'),
            3: dict(type='partition', message='no movement arming'),
            4: dict(type='partition', message='partial arming'),
            5: dict(type='partition', message='one-touch arming'),
            6: dict(type='partition', message='Future use 1'),
            7: dict(type='partition', message='Future use 2'),
            8: dict(type='partition', message='(InTouch) voice module arming'),
            9: dict(type='partition', message='Delinquency c;psomg'),
            10: dict(type='partition', message='Future use 3'),
        }),
    13: dict(
        level=EventLevel.INFO, tags=['disarm'], change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type='partition', message='{label} disarmed with master'),
    14: dict(
        level=EventLevel.INFO, tags=['disarm'], change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type='partition', message='{label} disarmed with user code'),
    15: dict(
        level=EventLevel.INFO, tags=['disarm'], change=dict(arm=False, current_state="disarmed", audible_alarm=False),
        type='partition', message='{label} disarmed with keyswitch'),
    16: dict(
        level=EventLevel.INFO, tags=['disarm', 'after_alarm'],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False), type='partition',
        message='{label} disarmed after alarm with master'),
    17: dict(
        level=EventLevel.INFO, tags=['disarm', 'after_alarm'],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False), type='partition',
        message='{label} disarmed after alarm with user code'),
    18: dict(
        level=EventLevel.INFO, tags=['disarm', 'after_alarm'],
        change=dict(arm=False, current_state="disarmed", audible_alarm=False), type='partition',
        message='{label} disarmed after alarm with keyswitch'),
    19: dict(
        level=EventLevel.INFO, tags=['alarm', 'cancel'], type='user', message='Alarm cancelled by {label} (master)'),
    20: dict(
        level=EventLevel.INFO, tags=['alarm', 'cancel'], type='user', message='Alarm cancelled by {label} (user code)'),
    21: dict(
        level=EventLevel.INFO, tags=['alarm', 'cancel'], type='keyswitch',
        message='{label} Alarm cancelled with keyswitch'),
    22: dict(
        level=EventLevel.INFO, tags=['disarm'], change=dict(arm=False), type='partition',
        message='{label} special disarming', sub={  # no label passed
            0: dict(type='partition', message='Auto arm cancelled'),
            1: dict(type='partition', message='One-touch stay/instant disarm'),
            2: dict(type='partition', message='Disarming with Winload by {@user}', user=_minor2_provider),
            3: dict(type='partition', tags=['after_alarm'], message='Disarming with Winload after alarm by {@user}',
                    user=_minor2_provider),
            4: dict(type='partition', message='Winload cancelled alarm by {@user}', user=_minor2_provider),
            5: dict(type='partition', message='Future use'),
            6: dict(type='partition', message='Future use'),
            7: dict(type='partition', message='Future use'),
            8: dict(type='partition', message='(InTouch) voice module disarming'),
            9: dict(type='partition', message='Future use')
        }),
    23: dict(
        level=EventLevel.INFO, type='zone', change=dict(bypassed=_toggle_provider), tags=['bypass'],
        message='Zone {label} bypass toggled'),
    24: dict(
        level=EventLevel.CRITICAL, tags=['alarm', 'trigger'], type='zone',
        change=dict(generated_alarm=True, presently_in_alarm=True), message='Zone {label} in alarm',
        hook_fn=_zone_generated_alarm),
    25: dict(
        level=EventLevel.CRITICAL, tags=['alarm', 'fire', 'trigger'], type='zone', change=dict(fire_alarm=True),
        message='Zone {label} in fire alarm', hook_fn=_request_status_refresh),
    26: dict(
        level=EventLevel.WARN, tags=['alarm', 'restore'], type='zone', change=dict(presently_in_alarm=False),
        message='Zone {label} alarm restore', hook_fn=_request_status_refresh),
    27: dict(
        level=EventLevel.WARN, tags=['alarm', 'fire', 'restore'], type='zone', change=dict(fire_alarm=False),
        message='Zone {label} fire alarm restore', hook_fn=_request_status_refresh),
    28: dict(
        level=EventLevel.INFO, type='user', message='{label} Early to disarm by user'),
    29: dict(
        level=EventLevel.INFO, type='user', message='{label} Late to disarm by user'),
    30: dict(
        level=EventLevel.CRITICAL, type='special', message='Special alarm', tags=['alarm', 'trigger'], sub={
            0: dict(message='Panic emergency'),
            1: dict(message='Panic medical'),
            2: dict(message='Panic fire'),
            3: dict(message='Recent closing'),
            4: dict(message='Police code'),
            5: dict(message='Zone shutdown'),
            6: dict(message='Future use 1'),
            7: dict(message='Future use 2'),
            8: dict(message='TLM alarm'),
            9: dict(message='Central communication failure alarm'),
            10: dict(message='Module tamper alarm'),
            11: dict(message='Missing GSM module alarm'),
            12: dict(message='GSM no service alarm'),
            13: dict(message='Missing IP module alarm'),
            14: dict(message='IP no service alarm'),
            15: dict(message='Missing voice module alarm'),
        }, hook_fn=_request_status_refresh),
    31: dict(
        level=EventLevel.CRITICAL, type='user', tags=['alarm', 'trigger'], message='Duress alarm by user {label}',
        hook_fn=_request_status_refresh),
    32: dict(
        level=EventLevel.INFO, type='zone', change=dict(shutdown=True), message='Zone {label} shutdown'),
    33: dict(
        level=EventLevel.INFO, tags=['tamper'], type='zone', change=dict(zone_tamper_trouble=True),
        message='Zone {label} tamper'),
    34: dict(
        level=EventLevel.INFO, tags=['tamper', 'restore'], type='zone', change=dict(zone_tamper_trouble=True),
        message='Zone {label} tamper restore'),
    35: dict(
        level=EventLevel.INFO, tags=['tamper'], type='zone', message='Zone special {label} tamper', sub={
            0: dict(message='Keypad Lockout'),
            1: dict(message='Voice lockout')
        }),
    36: dict(
        level=EventLevel.WARN, type='system', message='Trouble', tags=['trouble'], sub={
            0: dict(message='TLM trouble'),
            1: dict(message='AC failure'),
            2: dict(message='Battery failure'),
            3: dict(message='Auxiliary current overload'),
            4: dict(message='Bell current overload'),
            5: dict(message='Bell disconnected'),
            6: dict(tags=['clock'], message='Clock loss'),
            7: dict(message='Fire loop trouble'),
            8: dict(message='Fail to communicate to monitoring station telephone #1'),
            9: dict(message='Fail to communicate to monitoring station telephone #2'),
            11: dict(message='Fail to communicate to voice report'),
            12: dict(message='RF jamming'),
            13: dict(message='GSM RF jamming'),
            14: dict(message='GSM no service'),
            15: dict(message='GSM supervision lost'),
            16: dict(message='Fail To Communicate IP Receiver 1 (GPRS)'),
            17: dict(message='Fail To Communicate IP Receiver 2 (GPRS)'),
            18: dict(message='IP Module No Service'),
            19: dict(message='IP Module Supervision Loss'),
            20: dict(message='Fail To Communicate IP Receiver 1 (IP)'),
            21: dict(message='Fail To Communicate IP Receiver 2 (IP)')
        }),
    37: dict(
        level=EventLevel.INFO, type='system', message='Trouble restore', tags=['trouble', 'restore'], sub={
            0: dict(message='TLM trouble restore'),
            1: dict(message='AC failure restore'),
            2: dict(message='Battery failure restore'),
            3: dict(message='Auxiliary current overload restore'),
            4: dict(message='Bell current overload restore'),
            5: dict(message='Bell disconnected restore'),
            6: dict(tags=['clock'], message='Clock loss restore'),
            7: dict(message='Fire loop trouble restore'),
            8: dict(message='Fail to communicate to monitoring station telephone #1 restore'),
            9: dict(message='Fail to communicate to monitoring station telephone #2 restore'),
            11: dict(message='Fail to communicate to voice report restore'),
            12: dict(message='RF jamming restore'),
            13: dict(message='GSM RF jamming restore'),
            14: dict(message='GSM no service restore'),
            15: dict(message='GSM supervision lost restore'),
            16: dict(message='Fail To Communicate IP Receiver 1 (GPRS) restore'),
            17: dict(message='Fail To Communicate IP Receiver 2 (GPRS) restore'),
            18: dict(message='IP Module No Service restore'),
            19: dict(message='IP Module Supervision Loss restore'),
            20: dict(message='Fail To Communicate IP Receiver 1 (IP) restore'),
            21: dict(message='Fail To Communicate IP Receiver 2 (IP) restore'),
            99: dict(message='Any trouble event restore')
        }),
    38: dict(
        level=EventLevel.WARN, type='system', message='Module trouble', tags=['trouble'], sub={
            0: dict(message='Combus fault'),
            1: dict(message='Module tamper'),
            2: dict(message='ROM/RAM error'),
            3: dict(message='TLM trouble'),
            4: dict(message='Fail to communicate'),
            5: dict(message='Printer fault'),
            6: dict(message='AC failure'),
            7: dict(message='Battery failure'),
            8: dict(message='Auxiliary failure'),
            9: dict(message='Future use'),
        }, module=_minor2_provider),
    39: dict(
        level=EventLevel.INFO, type='system', message='Module trouble restore', tags=['trouble', 'restore'], sub={
            0: dict(message='Combus fault restored'),
            1: dict(message='Module tamper restored'),
            2: dict(message='ROM/RAM error restored'),
            3: dict(message='TLM trouble restored'),
            4: dict(message='Fail to communicate restored'),
            5: dict(message='Printer fault restored'),
            6: dict(message='AC failure restored'),
            7: dict(message='Battery failure restored'),
            8: dict(message='Auxiliary failure restored'),
            9: dict(message='Future use'),
        }, module=_minor2_provider),
    40: dict(
        level=EventLevel.WARN, type='system', message='Fail to communicate on telephone number',
        telephone=_minor2_provider),
    41: dict(
        level=EventLevel.WARN, type='zone', changes=dict(zone_low_battery_trouble=True), tags=['trouble'],
        message='Low battery on zone {label}'),
    42: dict(
        level=EventLevel.WARN, type='zone', changes=dict(zone_supervision_trouble=True), tags=['trouble'],
        message='Zone {label} supervision trouble'),
    43: dict(
        level=EventLevel.INFO, type='zone', changes=dict(zone_low_battery_trouble=False), tags=['trouble', 'restore'],
        message='Low battery on zone {label} restored'),
    44: dict(
        level=EventLevel.INFO, type='zone', changes=dict(zone_supervision_trouble=False), tags=['trouble', 'restore'],
        message='Zone {label} supervision trouble restored'),
    45: dict(
        level=EventLevel.INFO, type='system', message='Special events', sub={
            0: 'Power-up after total power down',
            1: 'Software reset (watchdog)',
            2: 'Test report',
            3: 'Listen-in request',
            4: 'WinLoad in (connected)',
            5: 'WinLoad out (disconnected)',
            6: 'Installer in programming',
            7: 'Installer out of programming',
            8: 'Future use'
        }),
    46: dict(
        level=EventLevel.WARN, type='user', message='Early to arm by user'),
    47: dict(
        level=EventLevel.WARN, type='user', message='Late to arm by user'),
    48: dict(
        level=EventLevel.INFO, type='utility', message='Utility key'),
    49: dict(
        level=EventLevel.WARN, type='door', message='Request for exit'),
    50: dict(
        level=EventLevel.WARN, type='door', message='Access denied'),
    51: dict(
        level=EventLevel.WARN, type='door', message='Door left open alarm'),
    52: dict(
        level=EventLevel.WARN, type='door', message='Door forced alarm'),
    53: dict(
        level=EventLevel.WARN, type='door', message='Door left open restore'),
    54: dict(
        level=EventLevel.WARN, type='door', message='Door forced open restore'),
    55: dict(
        level=EventLevel.WARN, type='zone', message='Intellizone triggered'),
    56: dict(
        level=EventLevel.WARN, type='zone', message='Zone excluded on Force arming'),
    57: dict(
        level=EventLevel.WARN, type='zone', message='Zone went back to arm status'),
    58: dict(
        level=EventLevel.INFO, type='module', message='New module assigned on combus'),
    59: dict(
        level=EventLevel.INFO, type='module', message='Module manually removed from combus'),
    60: dict(
        level=EventLevel.INFO, type='system', message='Non-saved event', sub={
            0: 'Remote control rejected',
            1: 'Future use'
        }),
    61: dict(
        level=EventLevel.INFO, type='module', message='Module manually removed from combus'),
    62: dict(
        level=EventLevel.INFO, type='user', message='Access granted to user'),
    63: dict(
        level=EventLevel.INFO, type='user', message='Access denied to user'),
    64: dict(
        level=EventLevel.INFO, type='system', message='Status 1', sub={
            0: 'Armed',
            1: 'Force armed',
            2: 'Stay armed',
            3: 'Instant armed',
            4: 'Strobe alarm',
            5: 'Silent alarm',
            6: 'Audible alarm',
            7: 'Fire alarm'
        }),
    65: dict(
        level=EventLevel.INFO, type='system', message='Status 2', sub={
            0: 'Ready',
            1: 'Exit delay',
            2: 'Entry delay',
            3: 'System in trouble',
            4: 'Alarm in memory',
            5: 'Zones bypassed',
            6: 'Bypass, master, installer programming',
            7: 'Keypad lockout'
        }),
    66: dict(
        level=EventLevel.INFO, type='system', message='Status 3', sub={
            0: 'Intellizone delay engaged',
            1: 'Fire delay engaged',
            2: 'Auto arm',
            3: 'Arming with voice module (set until exit delay finishes)',
            4: 'Tamper',
            5: 'Zone low battery',
            6: 'Fire loop trouble',
            7: 'Zone supervision trouble'
        }),
    67: dict(
        level=EventLevel.INFO, type='system', message='Special status', sub={
            # 0-3: Chime in partition 1-4
            4: 'Smoke detector power reset',
            5: 'Ground start',
            6: 'Kiss off',
            7: 'Telephone ring',
            # 8-15: Bell on partition 1-8
            # 16-23: Pulsed alarm in partition
            # 24-31: Open/close Kiss off in partition
            # 32-63: Keyswitch/PGM input
            # 64-95: Status of access door
            96: 'Trouble in system',
            97: 'Trouble in dialer',
            98: 'Trouble in module',
            99: 'Trouble in combus',
            103: 'Time and date trouble',
            104: 'AC failure',
            105: 'Battery failure',
            106: 'Auxiliary current limit',
            107: 'Bell current limit',
            108: 'Bell absent',
            109: 'ROM error',
            110: 'RAM error',
            111: 'Future use',
            112: 'TLM 1 trouble',
            113: 'Fail to communicate 1',
            114: 'Fail to communicate 2',
            115: 'Fail to communicate 3',
            116: 'Fail to communicate 4',
            117: 'Fail to communicate with PC',
            120: 'Module tamper trouble',
            121: 'Module ROM error',
            122: 'Module TLM error',
            123: 'Module Failure to communicate',
            124: 'Module printer trouble',
            125: 'Module AC failure',
            126: 'Module battery trouble',
            127: 'Module auxiliary failure',
            128: 'Missing  keypad',
            129: 'Missing  module',
            133: 'Global combus failure',
            134: 'Combus overload',
            136: 'Dialer relay',
        }),
}



from paradox.event import EventLevel

property_map = {
    "ac_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "AC Power failure",
                 "False": "AC Power restored"}),
    "aux_limit_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "{Type} {label} AUX Limit reached",
                 "False": "{Type} {label} AUX Limit is OK"}),
    "generated_alarm": dict(
        level=EventLevel.CRITICAL,  # Flag is set when alarm is triggered and unset when armed
        # next time
        message={"True": "{Type} {label} in alarm",
                 "False": "{Type} {label} not in alarm"}),
    "fire_alarm": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} in fire alarm",
                 "False": "{Type} {label} not in fire alarm"}),
    "alarm_duration_finished": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} alarm duration finished",
                 "False": "{Type} {label} alarm duration active"}),
    "alarm_in_memory": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} has alarm in memory",
                 "False": "{Type} {label} alarm memory empty"}),
    "arm": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} is armed",
                 "False": "{Type} {label} is disarmed"}),
    "arm_away": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} away arm active",
                 "False": "{Type} {label} away arm inactive"}),
    "arm_no_entry": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} no entry arm is active",
                 "False": "{Type} {label} no entry arm is inactive"}),
    "arm_stay": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} stay arm active",
                 "False": "{Type} {label} stay arm inactive"}),
    "all_zone_closed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} all zones closed",
                 "False": "{Type} {label} not all zones closed"}),
    "audible_alarm": dict(
        level=EventLevel.CRITICAL,  # Flag is set when partition is in alarm. Does not mean it
        # produces any sound at that moment. Unset when disarmed.
        message={"True": "{Type} {label} alarm is audible",
                 "False": "{Type} {label} alarm is silent"}),
    "auto_arm_reach": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} auto arm reached",
                 "False": "{Type} {label} auto arm not reached"}),
    "auto_arming_engaged": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} auto arming engaged",
                 "False": "{Type} {label} auto arming is not engaged"}),
    "bypass_ready": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} bypass is ready",
                 "False": "{Type} {label} bypass is not ready"}),
    "cancel_alarm_reporting_on_disarming": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],  #
        message={"True": "{Type} {label} alarm reporting on disarm is inactive",
                 "False": "{Type} {label} alarm reporting on disarm is active"}),
    "battery_failure_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Battery failure trouble",
                 "False": "Battery failure trouble recovered"}),
    "follow_become_delay": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} follow becode delay is active",
                 "False": "{Type} {label} follow becode delay is inactive"}),
    "bell_absent_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Bell is absent",
                 "False": "Bell is present"}),
    "bell_limit_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Bell output overloaded",
                 "False": "Bell output load is adequate"}),
    "bypassed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} is bypassed",
                 "False": "{Type} {label} is not bypassed"}),
    "force_ready": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} force ready",
                 "False": "{Type} {label} force not ready"}),
    "inhibit_ready": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} inhibit ready",
                 "False": "{Type} {label} inhibit not ready"}),
    "intellizone_engage": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} intellizone engadged",
                 "False": "{Type} {label} intellizone not engadged"}),
    "lockout": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "{Type} {label} lockout",
                 "False": "{Type} {label} lockout inactive"}),
    "no_movement_delay_end": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} no movement delay ended",
                 "False": "{Type} {label} no movement delay is active"}),
    "entry_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} in entry delay",
                 "False": "{Type} {label} not in entry delay"}),
    "entry_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} entry delay is finished",
                 "False": "{Type} {label} entry delay is not finished"}),
    "exit_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} in exit delay",
                 "False": "{Type} {label} not in exit delay"}),
    "exit_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} exit delay is finished",
                 "False": "{Type} {label} exit delay is not finished"}),
    "fire_delay_in_progress": dict(
        level=EventLevel.INFO,  #
        message={"True": "{Type} {label} is in fire delay",
                 "False": "{Type} {label} is in fire delay"}),
    "fire_delay_end": dict(
        level=EventLevel.INFO,  #
        message={"True": "{Type} {label} fire delay ended",
                 "False": "{Type} {label} fire delay not ended"}),
    "intellizone_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} intellizone delay finished",
                 "False": "{Type} {label} intellizone delay is active"}),
    "module_tamper_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Module tampered",
                 "False": "Module tamper cleared"}),
    "open": dict(
        level=EventLevel.DEBUG,
        message={"True": "Zone {label} open",
                 "False": "Zone {label} closed"}),
    "panic_alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm'],
        message={"True": "{Type} {label} panic alarm active",
                 "False": "{Type} {label} panic alarm inactive"}),
    "partition_recently_close": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} partition recently close active",
                 "False": "{Type} {label} partition recently close inactive"}),
    "programming": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} programming active",
                 "False": "{Type} {label} programming inactive"}),
    "ready": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} ready",
                 "False": "{Type} {label} not ready"}),
    "remote_arming": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} remote arming active",
                 "False": "{Type} {label} remote arming inactive"}),
    "police_code_delay": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} police code delay active",
                 "False": "{Type} {label} police code delay inactive"}),
    "silent_alarm": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} silent alarm is active",
                 "False": "{Type} {label} silent alarm is inactive"}),
    "stay_arming_auto": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} auto stay arming active",
                 "False": "{Type} {label} auto stay arming inactive"}),
    "stay_instant_ready": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} stay instant ready",
                 "False": "{Type} {label} stay instant not ready"}),
    "supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "{Type} {label} supervision trouble",
                 "False": "{Type} {label} supervision OK"}),
    "tamper": dict(
        level=EventLevel.WARN, tags=['trouble'],
        message={"True": "{Type} {label} tampered",
                 "False": "{Type} {label} tamper cleared"}),
    "time_to_refresh_zone_status": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} time to refresh zone status",
                 "False": "{Type} {label} time to refresh zone status not came"}),
    "trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "{Type} {label} has trouble",
                 "False": "No trouble for {Type} {label}"}),
    "tx_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} tx delay finished",
                 "False": "{Type} {label} tx delay not finished"}),
    "voice_arming": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} voice arming active",
                 "False": "{Type} {label} voice arming inactive"}),
    "was_in_alarm": dict(
        level=EventLevel.INFO,  # Flag is set when partition is in alarm, unset when disarmed
        message={"True": "{Type} {label} was in alarm",
                 "False": "{Type} {label} was not in alarm"}),
    "zone_bypassed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} a zone is bypassed",
                 "False": "{Type} {label} no zones bypassed"}),
    "zone_fire_loop_trouble": dict(
        level=EventLevel.WARN, tags=['trouble'],
        message={"True": "{Type} {label} zone fire loop trouble",
                 "False": "{Type} {label} zone fire loop OK"}),
    "zone_supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Zone with supervision trouble",
                 "False": "Zone supervision is OK"}),
    "zone_tamper_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Zone tampered",
                 "False": "No Zone is tampered"}),
    "zone_low_battery_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "zone low battery trouble",
                 "False": "zone low battery trouble recovered"}),
    "bus_com_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "bus com trouble",
                 "False": "bus com OK"}),
    "bus_global_fail": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "bus global fail",
                 "False": "bus OK"}),
    "bus_overload_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "bus overload trouble",
                 "False": "bus load OK"}),
    "com_pc_trouble": dict(
        level=EventLevel.INFO, tags=['trouble'],
        message={"True": "pc communication trouble",
                 "False": "pc communication OK"}),
    "dialer_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "dialer trouble active",
                 "False": "dialer trouble inactive"}),
    "fail_tel_1_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "fail tel 1 trouble active",
                 "False": "fail tel 1 trouble inactive"}),
    "fail_tel_2_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "fail tel 2 trouble active",
                 "False": "fail tel 2 trouble inactive"}),
    "fail_tel_3_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "fail tel 3 trouble active",
                 "False": "fail tel 3 trouble inactive"}),
    "fail_tel_4_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "fail tel 4 trouble active",
                 "False": "fail tel 4 trouble inactive"}),
    "mdl_com_error": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module communication error active",
                 "False": "module communication error"}),
    "missing_keypad_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "missing keypad",
                 "False": "keypad present"}),
    "missing_module_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "missing a module",
                 "False": "all modules present"}),
    "module_ac_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module AC trouble",
                 "False": "all module AC OK"}),
    "module_aux_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module aux trouble",
                 "False": "module aux OK"}),
    "module_battery_fail": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module battery fail",
                 "False": "module battery OK"}),
    "module_fail_to_com_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module fail to communicate trouble",
                 "False": "module communication OK"}),
    "module_printer_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module printer trouble",
                 "False": "module printer OK"}),
    "module_rom_error_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module rom error trouble",
                 "False": "module rom error trouble recovered"}),
    "module_tlm_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module tlm trouble",
                 "False": "module tlm trouble recovered"}),
    "module_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "module trouble",
                 "False": "module trouble recovered"}),
    "rom_error_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "rom error trouble",
                 "False": "rom error trouble recovered"}),
    "safety_mismatch_trouble": dict(
        level=EventLevel.INFO, tags=['trouble'],
        message={"True": "safety mismatch trouble",
                 "False": "safety mismatch trouble recovered"}),
    "system_trouble": dict(
        level=EventLevel.INFO, tags=['trouble'],
        message={"True": "system trouble",
                 "False": "system trouble recovered"}),
    "time_lost_trouble": dict(
        level=EventLevel.INFO, tags=['trouble'],
        message={"True": "time lost trouble",
                 "False": "time lost trouble recovered"}),
    "tlm_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "tlm trouble",
                 "False": "tlm trouble recovered"}),
    "zone_fault_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "zone fault trouble",
                 "False": "zone fault trouble recovered"}),
    "activated_entry_delay": dict(
        level=EventLevel.INFO, tags=[],
        message={"True": "{Type} {label} activated entry delay active",
                 "False": "{Type} {label} activated entry delay inactive"}),
    "activated_intellizone_delay": dict(
        level=EventLevel.INFO, tags=[],
        message={"True": "{Type} {label} activated intellizone delay active",
                 "False": "{Type} {label} activated intellizone delay inactive"}),
    "low_battery": dict(
        level=EventLevel.INFO, tags=[],
        message={"True": "{Type} {label} low battery",
                 "False": "{Type} {label} battery OK"}),
    "presently_in_alarm": dict(
        level=EventLevel.INFO, tags=[],  # Flag is set when zone is in alarm
        message={"True": "{Type} {label} triggered alarm",
                 "False": "{Type} {label} alarm restored"}),
    "shut_down": dict(
        level=EventLevel.INFO, tags=[],
        message={"True": "{Type} {label} shut down active",
                 "False": "{Type} {label} shut down inactive"}),
    "tx_delay": dict(
        level=EventLevel.INFO, tags=[],
        message={"True": "{Type} {label} tx delay active",
                 "False": "{Type} {label} tx delay inactive"}),
    "time": dict(
        level=EventLevel.DEBUG, tags=['system', 'time'],
        message="Panel time is {value}"),
    # ---------------------
    # Synthetic properties
    # ---------------------
    "current_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "triggered": "{Type} {label} triggered alarm",
            "pending": "{Type} {label} arming pending",
            "armed_home": "{Type} {label} armed stay",
            "armed_away": "{Type} {label} armed away",
            "armed_night": "{Type} {label} armed night",
            "disarmed": "{Type} {label} disarmed"
        }
    ),
    "target_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "armed_home": "{Type} {label} arm stay",
            "armed_away": "{Type} {label} arm away",
            "armed_night": "{Type} {label} arm night",
            "disarmed": "{Type} {label} disarm"
        }
    )
}



from paradox.event import EventLevel

property_map = {
    "ac_failure_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'power'],
        message={"True": "AC Power failure",
                 "False": "AC Power restored"}),
    "ac_loss_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'power'],
        message={"True": "{Type} {label} AC Power lost",
                 "False": "{Type} {label} AC Power is OK"}),
    "alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm'],
        message={"True": "{Type} {label} in alarm",
                 "False": "{Type} {label} not in alarm"}),
    "alarms_in_memory": dict(
        level=EventLevel.INFO, tags=['alarm'],
        message={"True": "{Type} {label} has alarms in memory",
                 "False": "{Type} {label} alarms memory empty"}),
    "arm": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} is armed",
                 "False": "{Type} {label} is disarmed"}),
    "arm_force": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} forced arm",
                 "False": "{Type} {label} arm not forced"}),
    "arm_sleep": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} sleep arm is active",
                 "False": "{Type} {label} sleep arm is inactive"}),
    "arm_stay": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} stay arm active",
                 "False": "{Type} {label} stay arm inactive"}),
    "arm_with_remote": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} is armed with remote",
                 "False": "{Type} {label} is not armed with remote"}),
    "audible_alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm'],
        message={"True": "{Type} {label} alarm is audible",
                 "False": "{Type} {label} alarm is silent"}),
    "auto_arming_engaged": dict(
        level=EventLevel.CRITICAL, tags=['arm'],
        message={"True": "{Type} {label} auto arming engaged",
                 "False": "{Type} {label} auto arming is not engaged"}),
    "auxiliary_output_overload_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'output'],
        message={"True": "Auxiliary output overloaded",
                 "False": "Auxiliary output load is adequate"}),
    'battery': dict(
        level=EventLevel.DEBUG, tags=['voltage', 'battery'],
        message="Battery voltage is {value}V"),
    "battery_failure_trouble": dict(
        level=EventLevel.CRITICAL, tags=['battery', 'trouble'],
        message={"True": "{Type} {label} battery is low",
                 "False": "{Type} {label} battery is OK"}),
    "bell_activated": dict(
        level=EventLevel.CRITICAL, tags=['bell', 'alarm'],
        message={"True": "{Type} {label} bell is active",
                 "False": "{Type} {label} bell is inactive"}),
    "bell_delay_finished": dict(
        level=EventLevel.CRITICAL, tags=['bell', 'alarm'],
        message={"True": "{Type} {label} bell delay is inactive",
                 "False": "{Type} {label} bell delay is active"}),
    "bell_output_disconnected_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'bell'],
        message={"True": "Bell output disconnected",
                 "False": "Bell output connected"}),
    "bell_output_overload_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'bell'],
        message={"True": "Bell output overloaded",
                 "False": "Bell output load is adequate"}),
    "bell_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'bell'],
        message={"True": "Bell trouble",
                 "False": "Bell restored"}),
    "bypassed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} bypassed",
                 "False": "{Type} {label} not bypassed"}),
    "central_1_reporting_ftc_indicator_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'central'],
        message={"True": "Central 1 reporting trouble",
                 "False": "Central 1 reporting OK"}),
    "central_2_reporting_ftc_indicator_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'central'],
        message={"True": "Central 2 reporting trouble",
                 "False": "Central 2 reporting OK"}),
    "communication_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'communication'],
        message={"True": "Communication trouble",
                 "False": "Communication restored"}),
    "computer_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'communication'],
        message={"True": "PC communication failed",
                 "False": "PC communication OK"}),
    'dc': dict(
        level=EventLevel.DEBUG, tags=['voltage', 'power'],
        message="DC voltage is {value}V"),
    "entry_delay": dict(
        level=EventLevel.INFO, tags=['alarm'],
        message={"True": "{Type} {label} in entry delay",
                 "False": "{Type} {label} not in entry delay"}),
    "entry_delay_finished": dict(
        level=EventLevel.DEBUG, tags=['alarm'],
        message={"True": "{Type} {label} entry delay is finished",
                 "False": "{Type} {label} entry delay is not finished"}),
    "exit_delay": dict(
        level=EventLevel.INFO, tags=['arm'],
        message={"True": "{Type} {label} in exit delay",
                 "False": "{Type} {label} not in exit delay"}),
    "exit_delay_finished": dict(
        level=EventLevel.DEBUG, tags=['arm'],
        message={"True": "{Type} {label} exit delay is finished",
                 "False": "{Type} {label} exit delay is not finished"}),
    "fire_delay": dict(
        level=EventLevel.INFO, tags=['alarm', 'fire'],
        message={"True": "{Type} {label} is in fire delay",
                 "False": "{Type} {label} is in fire delay"}),
    "fire": dict(
        level=EventLevel.CRITICAL, tags=['alarm', 'fire'],
        message={"True": "{Type} {label} fire",
                 "False": "{Type} {label} has no fire"}),
    "fire_loop_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "Fire loop",
                 "False": "Fire loop recovered"}),
    "in_remote_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} in remote delay",
                 "False": "{Type} {label} not in remote delay"}),
    "intellizone_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} in intellizone delay",
                 "False": "{Type} {label} not in intellizone delay"}),
    "intellizone_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} intellizone delay finished",
                 "False": "{Type} {label} intellizone delay is active"}),
    "in_tx_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} in TX delay",
                 "False": "{Type} {label} not in TX delay"}),
    "keypad_battery_failure_trouble": dict(
        level=EventLevel.CRITICAL, tags=['battery', 'trouble'],
        message={"True": "Keypad {l} battery is low",
                 "False": "Keypad {label} battery is OK"}),
    "low_battery_trouble": dict(
        level=EventLevel.CRITICAL, tags=['battery', 'trouble'],
        message={"True": "{Type} {label} Battery is low",
                 "False": "{Type} {label} Battery level is adequate"}),
    "module_supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['supervision', 'trouble'],
        message={"True": "Module supervision trouble",
                 "False": "Module supervision OK"}),
    "module_tamper_trouble": dict(
        level=EventLevel.CRITICAL, tags=['tamper', 'trouble'],
        message={"True": "Module tampered",
                 "False": "Module tamper cleared"}),
    "no_delay": dict(
        level=EventLevel.CRITICAL,
        message={"True": "{Type} {label} has no delay",
                 "False": "{Type} {label} has a delay active"}),
    "open": dict(
        level=EventLevel.DEBUG,
        message={"True": "Zone {label} open",
                 "False": "Zone {label} closed"}),
    "pager_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL, tags=['pager', 'communication', 'trouble'],
        message={"True": "Pager communication failed",
                 "False": "Pager communication OK"}),
    "paramedic_alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm', 'medic'],
        message={"True": "{Type} {label} paramedic alarm active",
                 "False": "{Type} {label} paramedic alarm is inactive"}),
    "power_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'power'],
        message={"True": "Power trouble",
                 "False": "Power restored"}),
    "pulse_fire_alarm": dict(
        level=EventLevel.CRITICAL, tags=['fire', 'alarm'],
        message={"True": "{Type} {label} fire alarm pulse is active",
                 "False": "{Type} {label} fire alarm pulse is inactive"}),
    "ready_status": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} ready",
                 "False": "{Type} {label} not ready"}),
    "recent_closing_delay": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} recent closing delay",
                 "False": "{Type} {label} no recent closing delay"}),
    "rf_interference_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'rf', 'interference'],
        message={"True": "RF Interference trouble",
                 "False": "RF Interference cleared"}),
    "rf_low_battery_trouble": dict(
        level=EventLevel.CRITICAL, tags=['trouble', 'rf', 'battery'],
        message={"True": "{Type} {label} RF battery low",
                 "False": "{Type} {label} RF battery OK"}),
    'noise_floor': dict(
        level=EventLevel.DEBUG, tags=['rf', 'interference'],
        message="RF Noise floor is {value}"),
    "rf_supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['rf', 'supervision', 'trouble'],
        message={"True": "{Type} {label} has supervision trouble",
                 "False": "{Type} {label} has no supervision trouble"}),
    "shutdown": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} is shutdown",
                 "False": "{Type} {label} is active"}),
    "signal_strength": dict(
        level=EventLevel.DEBUG, tags=['rf'],
        message="{Type} {label} signal strength {value}"),
    "silent_alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm'],
        message={"True": "{Type} {label} silent alarm is active",
                 "False": "{Type} {label} silent alarm is inactive"}),
    "squawk": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} Squawk On",
                 "False": "{Type} {label} Squawk Off"}),
    "stayd_mode_active": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} StayD mode is active",
                 "False": "{Type} {label} StayD mode is inactive"}),
    "strobe_alarm": dict(
        level=EventLevel.CRITICAL, tags=['alarm'],
        message={"True": "{Type} {label} Alarm Strobe is active",
                 "False": "{Type} {label} Alarm Strobe is inactive"}),
    "supervision_failure_trouble": dict(
        level=EventLevel.CRITICAL, tags=['supervision', 'trouble'],
        message={"True": "{Type} {label} supervision lost",
                 "False": "{Type} {label} supervision is OK"}),
    "supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['supervision', 'trouble'],
        message={"True": "{Type} {label} supervision trouble",
                 "False": "{Type} {label} supervision OK"}),
    "tamper": dict(
        level=EventLevel.CRITICAL, tags=['tamper', 'trouble'],
        message={"True": "{Type} {label} tampered",
                 "False": "{Type} {label} tamper cleared"}),
    "telephone_line": dict(
        level=EventLevel.CRITICAL, tags=['telephone', 'communication', 'trouble'],
        message={"True": "Telephone line trouble",
                 "False": "Telephone line OK"}),
    'timer_loss_trouble': dict(
        level=EventLevel.CRITICAL, tags=['timer', 'trouble'],
        message={"True": "Timer lost trouble",
                 "False": "Timer recovered"}),
    "transmission_delay_finished": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} Transmission delay finished",
                 "False": "{Type} {label} Transmission delay is active"}),
    'trouble': dict(
        level=EventLevel.CRITICAL, tags=['trouble'],
        message={"True": "{Type} {label} has trouble",
                 "False": "No trouble for {Type} {label}"}),
    'vdc': dict(
        level=EventLevel.DEBUG, tags=['voltage', 'power'],
        message="VDC voltage is {value}V"),
    "voice_fail_to_communicate_trouble": dict(
        level=EventLevel.CRITICAL, tags=['voice', 'communication', 'trouble'],
        message={"True": "Voice module communication failed",
                 "False": "Voice module communication OK"}),
    "wait_window": dict(
        level=EventLevel.DEBUG,
        message={"True": "{Type} {label} wait window is active",
                 "False": "{Type} {label} wait window is inactive"}),
    "was_bypassed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} was bypassed",
                 "False": "{Type} {label} was not bypassed"}),
    "was_in_alarm": dict(
        level=EventLevel.INFO, tags=['alarm'],
        message={"True": "{Type} {label} was in alarm",
                 "False": "{Type} {label} was not in alarm"}),
    "wireless_keypad_ac_trouble": dict(
        level=EventLevel.CRITICAL, tags=['ac', 'power', 'trouble'],
        message={"True": "Wireless Keypad AC power with trouble",
                 "False": "Wireless Keypad AC power OK"}),
    "wireless_keypad_battery_trouble": dict(
        level=EventLevel.CRITICAL, tags=['battery', 'trouble'],
        message={"True": "Wireless Keypad battery with trouble",
                 "False": "Wireless Keypad battery OK"}),
    "wireless_repeater_ac_loss_trouble": dict(
        level=EventLevel.CRITICAL, tags=['ac', 'power', 'trouble'],
        message={"True": "Wireless Repeater with AC power failure",
                 "False": "Wireless Repeater AC power is OK"}),
    "wireless_repeater_battery_trouble": dict(
        level=EventLevel.CRITICAL, tags=['battery', 'trouble'],
        message={"True": "Wireless Repeater battery with trouble",
                 "False": "Wireless Repeater battery OK"}),
    "zone_bypassed": dict(
        level=EventLevel.INFO,
        message={"True": "{Type} {label} is bypassed",
                 "False": "{Type} {label} is not bypassed"}),
    "zone_supervision_trouble": dict(
        level=EventLevel.CRITICAL, tags=['supervision', 'trouble'],
        message={"True": "Zone with supervision trouble",
                 "False": "Zone supervision is OK"}),
    "zone_tamper_trouble": dict(
        level=EventLevel.CRITICAL, tags=['tamper', 'trouble'],
        message={"True": "Zone tampered",
                 "False": "No Zone is tampered"}),
    "time": dict(
        level=EventLevel.DEBUG, tags=['system', 'time'],
        message="Panel time is {value}"),

    # ---------------------
    # Synthetic properties
    # ---------------------
    "current_state": dict(
        level=EventLevel.INFO,
        tags=[],
        message={
            "triggered": "{Type} {label} triggered alarm",
            "pending": "{Type} {label} arming pending",
            "armed_home": "{Type} {label} armed stay",
            "armed_away": "{Type} {label} armed away",
            "disarmed": "{Type} {label} disarmed"
        }
    )
}



# -*- coding: utf-8 -*-


import logging
import typing

import serial_asyncio

from .connection import Connection
from .protocols import SerialConnectionProtocol

logger = logging.getLogger('PAI').getChild(__name__)


class SerialCommunication(Connection):
    def __init__(self, on_message: typing.Callable[[bytes], None], port, baud=9600):
        super().__init__(on_message=on_message)
        self.port_path = port
        self.baud = baud
        self.connected_future = None

    def on_port_closed(self):
        logger.error('Connection to panel was lost')
        if not self.connected_future.done():
            self.connected_future.set_result(False)
        self.connected = False

    def on_port_open(self):
        logger.info('Serial port open')
        self.connected_future.set_result(True)
        self.connected = True

    def open_timeout(self):
        if self.connected_future.done():
            return

        logger.error("Serial Port Timeout")
        self.connected_future.set_result(False)
        self.connected = False

    def make_protocol(self):
        return SerialConnectionProtocol(self.on_message, self.on_port_open, self.on_port_closed)

    async def connect(self):
        logger.info("Connecting to serial port {}".format(self.port_path))

        self.connected_future = self.loop.create_future()
        self.loop.call_later(5, self.open_timeout)

        _, self.connection = await serial_asyncio.create_serial_connection(
            self.loop, self.make_protocol, self.port_path, self.baud
        )

        return await self.connected_future



from .panel import Panel


def create_panel(core, product_id=None) -> Panel:
    if product_id is None:
        return Panel(core, product_id, True)
    elif product_id == 'DIGIPLEX_EVO_48':
        from . import evo
        return evo.Panel_EVO48(core, product_id, True)
    elif product_id == 'DIGIPLEX_EVO_96':
        from . import evo
        return evo.Panel_EVO96(core, product_id, True)
    elif product_id == 'DIGIPLEX_EVO_192':
        from . import evo
        return evo.Panel_EVO192(core, product_id, True)
    elif product_id == 'DIGIPLEX_EVO_HD':
        from . import evo
        return evo.Panel_EVOHD(core, product_id, True)
    elif product_id in [
        'SPECTRA_SP4000', 'SPECTRA_SP5500', 'SPECTRA_SP6000', 'SPECTRA_SP7000',
        'SPECTRA_SP65',
        'MAGELLAN_MG5000', 'MAGELLAN_MG5050'
    ]:
        from . import spectra_magellan
        return spectra_magellan.Panel(core, product_id, False)
    else:
        raise NotImplementedError(
            "We are not sure what panel you have (product_id: {}). \
            Please create an issue. Maybe we can help you.".format(str(product_id))
        )




from collections.abc import Mapping

from construct import Struct, RawCopy, BitStruct, Const, Nibble, Flag, Int8ub, BitsInteger, Int16ub, \
    Checksum, Bytes, this, Default, Padding, Enum, Int24ub, ExprAdapter, Byte, obj_, Array, Computed, Subconstruct, \
    ValidationError, ExprSymmetricAdapter, Bitwise, BitsSwapped, Embedded, ByteSwapped

from .adapters import StatusFlags, ZoneFlags, DateAdapter, PartitionStatus, EventAdapter, \
    ZoneFlagBitStruct, DictArray, EnumerationAdapter
from ..common import CommunicationSourceIDEnum, ProductIdEnum, calculate_checksum, PacketLength, PacketChecksum

LoginConfirmationResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0x1, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)
            ),
            "length" / PacketLength(Int8ub),
            "result" / BitStruct(
                "_not_used0" / BitsInteger(4),
                "partition_2" / Flag,
                "_not_used1" / BitsInteger(3)
            ),
            "callback" / Int16ub
        )),
    "checksum" / PacketChecksum(Bytes(1)))

InitializeCommunication = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / Struct("command" / Const(0x00, Int8ub)),
            "module_address" / Default(Int8ub, 0x00),  # (00= panel/module)
            "_not_used0" / Padding(2),
            "product_id" / ProductIdEnum,
            "firmware" / Struct(
                "version" / Int8ub,
                "revision" / Int8ub,
                "build" / Int8ub),
            "panel_id" / Int16ub,
            "pc_password" / Default(Bytes(2), b'0000'),
            "modem_speed" / Bytes(1),
            "source_method" / Default(Enum(Int8ub,
                                           Winload_Connection=0x00,
                                           NEware_Connection=0x55), 0x00),
            "user_code" / Default(Int24ub, 0x000000),
            "serial_number" / Bytes(4),
            "system_options" / BitsSwapped(BitStruct(
                Embedded(Struct(  # EVO section data 3030
                    "pgm1_smoke" / Flag,
                    "no_bell_cut_off" / Flag,
                    "daylight_saving_time" / Flag,
                    "shabbat_feature" / Flag,
                    "battery_charge_current" / Flag,
                    "ac_failure_not_displayed_as_trouble" / Flag,
                    "clear_bell_limit_trouble" / Flag,
                    "combus_speed" / Flag,
                )),
                "partitions" / StatusFlags(8),  # EVO section data 3031
                "siren_output_partition" / StatusFlags(8),  # EVO section data 3032
                Embedded(Struct(  # EVO section data 3033
                    "multiple_actions_user_menu" / Flag,
                    "user_code_length_flexible" / Flag,
                    "user_code_length_6" / Flag,
                    "power_save_mode" / Flag,
                    "bypass_not_displayed_when_armed" / Flag,
                    "trouble_latch" / Flag,
                    "eol_resistor_on_harwire_zones" / Flag,
                    "atz" / Flag,
                )),
                Embedded(Struct(  # EVO section data 3034
                    "wireless_transmitter_supervision_options" / BitsInteger(2),
                    "generate_supervision_failure_on_bypassed_wireless_zone" / Flag,
                    "restrict_arming_on_wireless_transmitter_supervision_failure" / Flag,
                    "tamper_recognition_options" / BitsInteger(2),
                    "generate_tamper_if_detected_on_bypassed_zone" / Flag,
                    "restrict_arming_on_tamper" / Flag,
                )),
                Embedded(Struct(  # EVO section data 3035
                    "restrict_arming_on_ac_failure" / Flag,
                    "restrict_arming_on_battery_failure" / Flag,
                    "restrict_arming_on_bell_or_aux_failure" / Flag,
                    "restrict_arming_on_tlm_failure" / Flag,
                    "restrict_arming_on_module_troubles" / Flag,
                    "account_number_transmission" / Flag,
                    "transmit_zone_status_on_serial_port" / Flag,
                    "serial_port_baud_rate_57600" / Flag,
                )),
                Embedded(Struct(  # EVO section data 3036
                    "telephone_line_monitoring" / BitsInteger(2),
                    "dialer_reporting" / Flag,
                    "dialing_method" / Flag,
                    "pulse_ratio" / Flag,
                    "busy_tone_detection" / Flag,
                    "switch_to_pulse_dialing" / Flag,
                    "bell_siren_upon_communication_failure" / Flag,
                )),
                Embedded(Struct(  # EVO section data 3037
                    "call_back" / Flag,
                    "automatic_event_buffer_transmission" / Flag,
                    "autotest_report_transmission_options" / BitsInteger(2),
                    "keypad_beep_on_successful_arming_disarming_report" / Flag,
                    "alternate_dialing" / Flag,
                    "dial_tone_delay" / Flag,
                    "report_zone_restore" / Flag,
                )),
                Embedded(Struct(  # EVO section data 3038
                    "access_control_feature" / Flag,
                    "log_request_for_exit" / Flag,
                    "log_door_left_open_restore" / Flag,
                    "log_door_forced_restore" / Flag,
                    "bulglar_alarm_on_forced_door" / Flag,
                    "skip_exit_delay_when_arming_with_access_card" / Flag,
                    "bulglar_alarm_on_door_left_open" / Flag,
                    "who_has_access_during_clock_loss" / Flag,
                )),
            )),
            "_not_used1" / Padding(4),
            "source_id" / Default(CommunicationSourceIDEnum, 1),
            "carrier_length" / Bytes(1)
        )),
    "checksum" / PacketChecksum(Bytes(1)))

RAMDataParserMap = {
    1: Struct(
        "_weekday" / Int8ub,
        "pgm_flags" / BitStruct(  # TODO: Do we need BitsSwapped here?
            'chime_zone_partition' / BitsSwapped(StatusFlags(4)),
            'power_smoke' / Flag,
            'ground_start' / Flag,
            'kiss_off' / Flag,
            'line_ring' / Flag,

            'bell_partition' / BitsSwapped(StatusFlags(8)),

            'fire_alarm' / BitsSwapped(StatusFlags(8)),

            'open_close_kiss_off' / BitsSwapped(StatusFlags(8)),
        ),
        "key-switch_triggered" / BitsSwapped(Bitwise(StatusFlags(32))),
        "door_open" / BitsSwapped(Bitwise(StatusFlags(32))),
        "system" / Struct(
            "troubles" / BitStruct(
                "system_trouble" / Flag,
                "dialer_trouble" / Flag,
                "module_trouble" / Flag,
                "bus_com_trouble" / Flag,  # BusCom
                "zone_tamper_trouble" / Flag,
                "zone_low_battery_trouble" / Flag,
                "zone_fault_trouble" / Flag,
                "time_lost_trouble" / Flag,

                "ac_trouble" / Flag,
                "battery_failure_trouble" / Flag,
                "aux_limit_trouble" / Flag,
                "bell_limit_trouble" / Flag,
                "bell_absent_trouble" / Flag,
                "rom_error_trouble" / Flag,
                "_future_use_0" / Flag,
                "_future_use_1" / Flag,

                "tlm_trouble" / Flag,
                "fail_tel_1_trouble" / Flag,
                "fail_tel_2_trouble" / Flag,
                "fail_tel_3_trouble" / Flag,
                "fail_tel_4_trouble" / Flag,
                "com_pc_trouble" / Flag,
                "_future_use_2" / Flag,
                "_future_use_3" / Flag,

                "module_tamper_trouble" / Flag,
                "module_rom_error_trouble" / Flag,
                "module_tlm_trouble" / Flag,
                "module_fail_to_com_trouble" / Flag,
                "module_printer_trouble" / Flag,
                "module_ac_trouble" / Flag,
                "module_battery_fail" / Flag,
                "module_aux_trouble" / Flag,

                "missing_keypad_trouble" / Flag,
                "missing_module_trouble" / Flag,
                "_future_use_4" / Flag,
                "_future_use_5" / Flag,
                "safety_mismatch_trouble" / Flag,
                "bus_global_fail" / Flag,
                "bus_overload_trouble" / Flag,
                "mdl_com_error" / Flag
            ),
            "date" / Struct(
                "weekday" / Computed(lambda ctx: ctx._._._weekday),
                "time" / DateAdapter(Bytes(7))
            ),
            "power" / Struct(
                "vdc" / ExprAdapter(Byte, lambda obj, ctx: round(obj * (20.3 - 1.4) / 255.0 + 1.4, 1), 0),
                "battery" / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
                "dc" / ExprAdapter(Byte, lambda obj, ctx: round(obj * 22.8 / 255.0, 1), 0),
            )
        ),
        "zone_open" / BitsSwapped(Bitwise(StatusFlags(96))),
        "zone_tamper" / BitsSwapped(Bitwise(StatusFlags(96))),
        "zone_low_battery" / BitsSwapped(Bitwise(StatusFlags(96))),
    ),
    2: Struct(
        "zone_status" / ZoneFlags(64)
    ),
    3: Struct(
        "zone_status" / ZoneFlags(32, start_index_from=65),
        "partition_status" / PartitionStatus(Bytes(32)),
    ),
    4: Struct(
        "partition_status" / PartitionStatus(Bytes(16)),
        "system" / Struct(
            "panel_status" / BitStruct(
                "installer_lock_active" / Flag,
                "_free" / Padding(7)
            ),
            "event" / Struct(
                "_event_pointer" / Int16ub,
                "_event_pointer_bus" / Int16ub,
            ),
            "_recycle_system" / Array(8, Int8ub),
            "report" / Struct(
                "arm_disarm_delay_timer" / Int8ub,
            )
        ),
        "_free" / Padding(34)
    ),
    5: Struct(
        "_free" / Padding(1),
        "bus-module_trouble" / BitsSwapped(Bitwise(StatusFlags(504)))
    )
}

DefinitionsParserMap = {
    "zone": BitStruct(
        "definition" / Enum(
            Nibble,
            disabled=0x0,
            entry_delay1=0x1,
            entry_delay2=0x2,
            follow=0x3,
            instant=0x4,
            buzzer_24h=0x5,
            burglary_24h=0x6,
            holdup_24h=0x7,
            gas_24h=0x8,
            heat_24h=0x9,
            water_24h=0xa,
            freeze_24h=0xb,
            delayed_fire_24h=0xc,
            standard_fire_24h=0xd,
            stay_delay1=0xe,
            stay_delay2=0xf,
        ),
        "partition" / Nibble,
        "options" / ByteSwapped(Struct(
            "auto_zone_shutdown_enabled" / Flag,
            "bypass_enabled" / Flag,
            "stay_zone" / Flag,
            "force_zone" / Flag,
            "alarm_type" / Enum(
                BitsInteger(2),
                steady_alarm=0x0,
                pulsed_alarm=0x1,
                silent_alarm=0x2,
                report_only=0x3,
            ),
            "intellizone" / Flag,
            "delay_before_transmission" / Flag,
        ))
    ),
    "partition": BitsSwapped(
        Bitwise(
            DictArray(8, 1, Struct(
                "_index" / Computed(this._index + 1),
                "definition" / ExprAdapter(
                    Default(Flag, False),
                    lambda obj, context: "enabled" if obj else "disabled",
                    lambda obj, context: obj == "enabled"
                )
            ))
        )),
}

LiveEvent = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0xE, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "event_source" / Const(0xFF, Int8ub),
            "event_nr" / Int16ub,
            "time" / DateAdapter(Bytes(6)),
            "event" / EventAdapter(Bytes(4)),
            "partition" / Computed(this.event.partition),
            "module_serial" / Bytes(4),
            "label_type" / Bytes(1),
            "label" / Bytes(16),
            "_not_used0" / Bytes(1),
        )), "checksum" / PacketChecksum(Bytes(1))
)

# "Event 1 requested (in compressed format) 1 (12 bytes)
#   Byte 00: [7-3]: Day, [2-0]: Month (MSB)
#   Byte 01: [7]: Month (LSB), [6-0]: Century
#   Byte 02: [7-1]: Year, [0]: Hour (MSB)
#   Byte 03: [7-4]: Hour (LSB), [3-0]: Minutes (MSB)
#   Byte 04: [7-6]: Minutes (LSB), [5-0]: Event Group
#   Byte 05: [7-4]: Partition, [3-0]: Event Number High Nibble
#   Byte 06: Event Number 1
#   Byte 07: Event Number 2
#   Byte 08: Serial Number 1 & 2 (2 nibbles)
#   Byte 09: Serial Number 3 & 4 (2 nibbles)
#   Byte 10: Serial Number 5 & 6 (2 nibbles)
#   Byte 11: Serial Number 7 & 8 (2 nibbles)"

CompressedEvent = Struct(
    "compressed" / BitStruct(
        "day" / BitsInteger(5),
        "month" / BitsInteger(4),
        "century" / BitsInteger(7),
        "year" / BitsInteger(7),
        "hour" / BitsInteger(5),
        "minute" / BitsInteger(6),
        "event_group" / BitsInteger(6),
        "partition" / BitsInteger(4),
        "event_1_high_nibble" / BitsInteger(2),
        "event_2_high_nibble" / BitsInteger(2)
    ),
    "minor_1" / Int8ub,
    "minor_2" / Int8ub,
    "module_serial" / Bytes(4)
)

RequestedEvent = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0xE, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "length" / PacketLength(Int8ub),
            "_not_used0" / Bytes(1),
            "requested_event_nr" / Const(0x00, Int8ub),
            "event_nr" / Int16ub,
            # "data" / Bytes(lambda this: this.length - 7)
            "data" / Array(lambda x: int((x.length - 7) / 12), CompressedEvent)
        )), "checksum" / Checksum(
        Bytes(1), lambda data: calculate_checksum(data), this.fields.data))

CloseConnection = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / Struct(
                "command" / Const(0x70, Int8ub)
            ),
            "length" / PacketLength(Int8ub),
            "message" / Default(Enum(
                Int8ub,
                requested_command_failed=0x00,
                invalid_user_code=0x01,
                partition_in_code_lockout=0x02,
                panel_will_disconnect=0x05,
                panel_not_connected=0x10,
                panel_already_connected=0x11,
                invalid_pc_password=0x12,
                winload_on_phone_line=0x13,
                invalid_module_address=0x14,
                cannot_write_in_ram=0x15,
                upgrade_request_fail=0x16,
                record_number_out_of_range=0x17,
                invalid_record_type=0x19,
                multibus_not_supported=0x1a,
                incorrect_number_of_users=0x1b,
                invalid_label_number=0x1c
            ), 0x05),
        )),
    "checksum" / PacketChecksum(Bytes(1)))


class EvoEEPROMAddressAdapter(Subconstruct):
    def deep_update(self, d, u):
        for k, v in u.items():
            if isinstance(v, Mapping):
                d[k] = self.deep_update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    def _build(self, obj, stream, context, path):
        if "control" in obj and obj["control"].get("ram_access"):  # for ram block
            ram_block = (obj["address"] & 0xf0000) >> 16
            self.deep_update(obj, dict(po=dict(block=ram_block)))
        else:  # for eeprom
            if obj["address"] >> 16 > 0x3:
                raise ValidationError("EEPROM address is out of range")
            eeprom_address_high_bits = (obj["address"] & 0x30000) >> 16
            self.deep_update(obj, dict(control=dict(eeprom_address_bits=eeprom_address_high_bits)))
        return self.subcon._build(obj, stream, context, path)

    def _parse(self, stream, context, path):
        obj = self.subcon._parsereport(stream, context, path)
        return obj


ReadEEPROM = Struct(
    "fields" / RawCopy(
        EvoEEPROMAddressAdapter(Struct(
            "po" / BitStruct(
                "command" / Const(0x5, Nibble),
                "block" / Default(Nibble, 0),
            ),
            "packet_length" / PacketLength(Int8ub),
            "control" / BitStruct(
                "ram_access" / Default(Flag, False),
                "alarm_reporting_pending" / Default(Flag, False),
                "Winload_connected" / Default(Flag, False),
                "NeWare_connected" / Default(Flag, False),
                "_not_used" / Default(BitsInteger(2), 0),
                "eeprom_address_bits" / Default(BitsInteger(2), 0)
            ),
            "bus_address" / Default(Int8ub, 0x00),  # 00 - Panel, 01-FF - Modules
            "address" / ExprSymmetricAdapter(Int16ub, obj_ & 0xffff),
            "length" / Int8ub
        ))),
    "checksum" / PacketChecksum(Bytes(1)))

ReadEEPROMResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0x5, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)
            ),
            "packet_length" / PacketLength(Int8ub),
            "control" / BitStruct(
                "ram_access" / Flag,  # RAM = 0 or EEPROM = 1
                "_not_used" / Padding(5),
                "eeprom_address_bits" / BitsInteger(2)  # EEPROM address bit 17 and 16
            ),
            "bus_address" / Int8ub,  # 00 - Panel, 01-FE - Modules
            "address" / Int16ub,
            "data" / Bytes(lambda x: x.packet_length - 7)
        )),
    "checksum" / PacketChecksum(Bytes(1)))

SetTimeDate = Struct(
    "fields" / RawCopy(Struct(
        "po" / Struct(
            "command" / Const(0x30, Int8ub)),
        "packet_length" / PacketLength(Int8ub),
        "_not_used0" / Padding(4),
        "century" / Int8ub,
        "year" / Int8ub,
        "month" / Int8ub,
        "day" / Int8ub,
        "hour" / Int8ub,
        "minute" / Int8ub,
    )),
    "checksum" / PacketChecksum(Bytes(1)))

SetTimeDateResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0x3, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "length" / Int8ub,
            "_not_used0" / Padding(4),
        )),
    "checksum" / PacketChecksum(Bytes(1)))

_PartitionCommandEnum = Enum(
    Nibble,
    none=0,
    arm=2,
    arm_stay=3,
    arm_instant=4,
    arm_force=5,
    disarm=6,
    beep_keypads=8
)

PerformPartitionAction = Struct(
    "fields" / RawCopy(Struct(
        "po" / Struct(
            "command" / Const(0x40, Int8ub)),
        "packet_length" / PacketLength(Int8ub),
        "_not_used0" / Padding(4),
        "partitions" / Bitwise(
            DictArray(8, 1, Struct(
                "_index" / Computed(this._index + 1),
                "command" / Default(_PartitionCommandEnum, "none")
            ), pick_key="command")
        ),
        "_not_used1" / Padding(4),
    )),
    "checksum" / PacketChecksum(Bytes(1)))

# Used for partitions and PGMs
PerformActionResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0x4, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "packet_length" / PacketLength(Int8ub),
            "_not_used0" / Padding(4),
        )),
    "checksum" / PacketChecksum(Bytes(1)))

ZoneActionBitOperation = Enum(
    Int8ub,
    set=0x08,
    clear=0x00
)

PerformZoneAction = Struct(
    "fields" / RawCopy(Struct(
        "po" / Struct(
            "command" / Const(0xd0, Int8ub)),
        "packet_length" / PacketLength(Int8ub),
        "flags" / ZoneFlagBitStruct,
        "operation" / ZoneActionBitOperation,
        "_not_used" / Padding(2),
        "zones" / BitsSwapped(Bitwise(
            EnumerationAdapter(Array(192, Flag)))),
    )),
    "checksum" / PacketChecksum(Bytes(1)))

PerformZoneActionResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0xd, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "packet_length" / PacketLength(Int8ub),
            "flags" / ZoneFlagBitStruct,
            "operation" / ZoneActionBitOperation,
            "_not_used" / Padding(2),
        )),
    "checksum" / PacketChecksum(Bytes(1)))

_PGMCommandEnum = Enum(
    Int8ub,
    release=0,
    off=1,
    on=3,
    on_override=4,
    off_override=2
)

PerformPGMAction = Struct(
    "fields" / RawCopy(Struct(
        "po" / Struct(
            "command" / Const(0x40, Int8ub)),
        "packet_length" / PacketLength(Int8ub),
        "unknown0" / Const(0x06, Int8ub),
        "_not_used0" / Padding(3),
        "pgms" / BitsSwapped(Bitwise(EnumerationAdapter(Array(8, Flag)))),
        "_not_used1" / Padding(7),
        "command" / _PGMCommandEnum,
        "_not_used1" / Padding(3),
    )),
    "checksum" / PacketChecksum(Bytes(1)))

ErrorMessage = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(0x7, Nibble),
                "status" / Struct(
                    "reserved" / Flag,
                    "alarm_reporting_pending" / Flag,
                    "Winload_connected" / Flag,
                    "NeWare_connected" / Flag)),
            "length" / PacketLength(Int8ub),
            "message" / Enum(Int8ub,
                             requested_command_failed=0x00,
                             invalid_user_code=0x01,
                             partition_in_code_lockout=0x02,
                             panel_will_disconnect=0x05,
                             panel_not_connected=0x10,
                             panel_already_connected=0x11,
                             invalid_pc_password=0x12,
                             winload_on_phone_line=0x13,
                             invalid_module_address=0x14,
                             cannot_write_in_ram=0x15,
                             upgrade_request_fail=0x16,
                             record_number_out_of_range=0x17,
                             invalid_record_type=0x19,
                             multibus_not_supported=0x1a,
                             incorrect_number_of_users=0x1b,
                             invalid_label_number=0x1c
                             ),
        )),
    "checksum" / PacketChecksum(Bytes(1)))



import json
import logging
import os
import typing
from collections import namedtuple

from paho.mqtt.client import MQTTMessage, Client

from paradox.config import config as cfg
from paradox.event import EventLevel, Event, Change, Notification
from paradox.lib import ps
from paradox.lib.utils import JSONByteEncoder, sanitize_key, call_soon_in_main_loop
from .core import AbstractMQTTInterface, ELEMENT_TOPIC_MAP

logger = logging.getLogger('PAI').getChild(__name__)

ParsedMessage = namedtuple('parsed_message', 'topics element content')


def mqtt_handle_decorator(func: typing.Callable[
    ["BasicMQTTInterface", ParsedMessage],
    typing.Coroutine[None, "BasicMQTTInterface", ParsedMessage]
]):
    def wrapper(self: "BasicMQTTInterface", client: Client, userdata, message: MQTTMessage):
        try:
            if message.retain:
                logger.warning(
                    "Ignoring command: retained message topic={}, payload={}".format(
                        message.topic, str(message.payload.decode("utf-8"))
                    )
                )
                return

            if self.alarm is None:
                logger.warning(
                    "No alarm. Ignoring command: message topic={}, payload={}".format(
                        message.topic, str(message.payload.decode("utf-8"))
                    )
                )
                return

            logger.info(
                "message topic={}, payload={}".format(
                    message.topic,
                    str(message.payload.decode("utf-8"))
                )
            )

            topic = message.topic.split(cfg.MQTT_BASE_TOPIC)[1]

            topics = topic.split("/")

            if len(topics) < 3:
                logger.error(
                    "Invalid topic in mqtt message: {}".format(message.topic))
                return

            content = message.payload.decode("utf-8").strip()

            element = None
            if len(topics) >= 4:
                element = topics[3]

            call_soon_in_main_loop(func(self, ParsedMessage(topics, element, content)))
        except Exception:
            logger.exception("Failed to execute command")

    return wrapper


class BasicMQTTInterface(AbstractMQTTInterface):
    def __init__(self, alarm):
        super().__init__(alarm)

        self.partitions = {}
        self.definitions = {}
        self.labels = {}

    def on_connect(self, mqttc, userdata, flags, result):
        ps.subscribe(self._handle_panel_labels, "labels_loaded")
        ps.subscribe(self._handle_panel_definitions, "definitions_loaded")
        ps.subscribe(self._handle_panel_change, "changes")
        ps.subscribe(self._handle_panel_event, "events")
        ps.subscribe(self._handle_connected, "connected")

        self.subscribe_callback(
            "{}/{}/{}/#".format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_CONTROL_TOPIC, cfg.MQTT_OUTPUT_TOPIC),
            self._mqtt_handle_output_control
        )
        self.subscribe_callback(
            "{}/{}/{}/#".format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_CONTROL_TOPIC, cfg.MQTT_ZONE_TOPIC),
            self._mqtt_handle_zone_control
        )
        self.subscribe_callback(
            "{}/{}/{}/#".format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_CONTROL_TOPIC, cfg.MQTT_PARTITION_TOPIC),
            self._mqtt_handle_partition_control
        )
        self.subscribe_callback(
            "{}/{}/{}".format(cfg.MQTT_BASE_TOPIC, cfg.MQTT_NOTIFICATIONS_TOPIC, "#"),
            self._mqtt_handle_notifications
        )

    @mqtt_handle_decorator
    async def _mqtt_handle_notifications(self, prep: ParsedMessage):
        topics = prep.topics
        try:
            level = EventLevel.from_name(topics[2].upper())
        except Exception as e:
            logger.error(e)
            return

        ps.sendNotification(Notification(sender=self.name, message=prep.content, level=level))

    @mqtt_handle_decorator
    async def _mqtt_handle_zone_control(self, prep: ParsedMessage):
        topics, element, command = prep
        if not await self.alarm.control_zone(element, command):
            logger.warning("Zone command refused: {}={}".format(element, command))

    @mqtt_handle_decorator
    async def _mqtt_handle_partition_control(self, prep: ParsedMessage):
        topics, element, command = prep
        command = cfg.MQTT_COMMAND_ALIAS.get(command, command)

        if command.startswith('code_toggle-'):
            tokens = command.split('-')
            if len(tokens) < 2:
                logger.warning("Invalid token length {}".format(len(tokens)))
                return

            if tokens[1] not in cfg.MQTT_TOGGLE_CODES:
                logger.warning("Invalid toggle code {}".format(tokens[1]))
                return

            if element.lower() == 'all':
                command = 'arm'

                for k, v in self.partitions.items():
                    # If "all" and a single partition is armed, default is
                    # to disarm
                    for k1, v1 in self.partitions[k].items():
                        if (k1 == 'arm' or k1 == 'exit_delay' or k1 == 'entry_delay') and v1:
                            command = 'disarm'
                            break

                    if command == 'disarm':
                        break

            elif element in self.partitions:
                if ('arm' in self.partitions[element] and self.partitions[element]['arm']) \
                        or ('exit_delay' in self.partitions[element] and self.partitions[element]['exit_delay']):
                    command = 'disarm'
                else:
                    command = 'arm'
            else:
                logger.warning("Element {} not found".format(element))
                return

            ps.sendNotification(
                Notification(
                    sender="mqtt",
                    message="Command by {}: {}".format(cfg.MQTT_TOGGLE_CODES[tokens[1]], command),
                    level=EventLevel.INFO
                )
            )

        logger.info("Partition command: {} = {}".format(element, command))
        if not await self.alarm.control_partition(element, command):
            logger.warning("Partition command refused: {}={}".format(element, command))

    @mqtt_handle_decorator
    async def _mqtt_handle_output_control(self, prep: ParsedMessage):
        topics, element, command = prep
        logger.debug("Output command: {} = {}".format(element, command))

        if not await self.alarm.control_output(element, command):
            logger.warning("Output command refused: {}={}".format(element, command))

    def _handle_panel_event(self, event: Event):
        """
        Handle Live Event

        :param raw: object with properties (can have byte properties)
        :return:
        """

        if cfg.MQTT_PUBLISH_RAW_EVENTS:
            self.publish(
                '{}/{}/{}'.format(
                    cfg.MQTT_BASE_TOPIC,
                    cfg.MQTT_EVENTS_TOPIC,
                    cfg.MQTT_RAW_TOPIC
                ),
                json.dumps(event.props, ensure_ascii=False, cls=JSONByteEncoder, default=str, sort_keys=True),
                0, cfg.MQTT_RETAIN
            )

    def _handle_panel_labels(self, data: dict):
        self.labels = data

    def _handle_panel_definitions(self, data: dict):
        self.definitions = data

    def _handle_connected(self):
        # After we get 2 partitions, lets publish a dashboard
        if cfg.MQTT_DASH_PUBLISH and len(self.partitions) == 2:
            self._publish_dash(cfg.MQTT_DASH_TEMPLATE, list(self.partitions.keys()))

        for element_type in self.definitions:  # zones, partitions
            labels = self.labels[element_type]
            definitions = self.definitions[element_type]
            for i in definitions:  # numeric index
                if i not in labels:
                    continue

                for attribute in definitions[i]:  # attribute
                    self._publish(f'{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_DEFINITION_TOPIC}',
                                  element_type,
                                  labels[i]['key'],
                                  attribute,
                                  definitions[i][attribute])

    def _handle_panel_change(self, change: Change):
        attribute = change.property
        label = change.key
        value = change.new_value
        element_type = change.type

        """Handle Property Change"""
        if label not in self.partitions:
            self.partitions[label] = dict()

        self.partitions[label][attribute] = value
        self._publish(f'{cfg.MQTT_BASE_TOPIC}/{cfg.MQTT_STATES_TOPIC}', element_type, label, attribute, value)

    def _publish(self, base: str, element_type: str, label: str, attribute: str, value: [str, int, bool]):
        if element_type in ELEMENT_TOPIC_MAP:
            element_topic = ELEMENT_TOPIC_MAP[element_type]
        else:
            element_topic = element_type

        if isinstance(value, dict):
            # This is fragile...
            if '/' in attribute and not attribute.startswith('/'):
                attribute = f"/{attribute}"

            for attr_name, attr_value in value.items():
                label_tp = f"{attribute}/{attr_name}"
                self._publish(base, element_type, label, label_tp, attr_value)
            return

        if cfg.MQTT_USE_NUMERIC_STATES:
            publish_value = int(value)
        else:
            publish_value = value

        self.publish(
            '{}/{}/{}/{}'.format(
                base,
                element_topic,
                sanitize_key(label),
                attribute
            ),
            "{}".format(publish_value), 0, cfg.MQTT_RETAIN
        )

    def _publish_dash(self, fname, partitions):
        # TODO: move to a separate component
        if len(partitions) < 2:
            return

        if os.path.exists(fname):
            with open(fname, 'r') as f:
                data = f.read()
                data = data.replace('__PARTITION1__', partitions[0]).replace('__PARTITION2__', partitions[1])
                self.publish(cfg.MQTT_DASH_TOPIC, data, 2, True)
                logger.info("MQTT Dash panel published to {}".format(cfg.MQTT_DASH_TOPIC))
        else:
            logger.warning("MQTT DASH Template not found: {}".format(fname))



# -*- coding: utf-8 -*-

import asyncio
import logging
import time
from binascii import hexlify
from typing import Optional, Sequence, Iterable, Callable

from construct import Container

from paradox.data.enums import RunState
from paradox.event import Event, LiveEvent, ChangeEvent, Change
from paradox.config import config as cfg
from paradox.connections.ip_connection import IPConnection
from paradox.connections.serial_connection import SerialCommunication
from paradox.data.memory_storage import MemoryStorage as Storage
from paradox.exceptions import StatusRequestException
from paradox.hardware import create_panel
from paradox.lib import ps
from paradox.lib.async_message_manager import EventMessageHandler, ErrorMessageHandler
from paradox.lib.utils import deep_merge
from paradox.parsers.status import convert_raw_status

logger = logging.getLogger('PAI').getChild(__name__)


def async_loop_unhandled_exception_handler(loop, context):
    logger.error("Unhandled exception in async loop(%s): %s", loop, context)

    loop.default_exception_handler(context)
    # exception = context.get('exception')
    logger.exception("Unhandled exception in async loop")


class Paradox:
    def __init__(self, retries=3):
        self.panel = None  # type: Panel
        self._connection = None
        self.retries = retries
        self.work_loop = asyncio.get_event_loop()  # type: asyncio.AbstractEventLoop
        self.work_loop.set_exception_handler(async_loop_unhandled_exception_handler)
        self.receive_worker_task = None

        self.storage = Storage()

        self._run_state = RunState.STOP
        self.request_lock = asyncio.Lock()
        self.busy = asyncio.Lock()
        self.loop_wait_event = asyncio.Event()

        ps.subscribe(self._on_labels_load, "labels_loaded")
        ps.subscribe(self._on_definitions_load, "definitons_loaded")
        ps.subscribe(self._on_status_update, "status_update")
        ps.subscribe(self._on_event, "events")
        ps.subscribe(self._on_property_change, "changes")

    @property
    def run_state(self):
        return self._run_state

    @run_state.setter
    def run_state(self, value: RunState):
        self._run_state = value
        ps.sendMessage("run-state", state=value)

    @property
    def connection(self):
        if not self._connection:
            # Load a connection to the alarm
            if cfg.CONNECTION_TYPE == "Serial":
                logger.info("Using Serial Connection")

                self._connection = SerialCommunication(
                    self.on_connection_message, port=cfg.SERIAL_PORT, baud=cfg.SERIAL_BAUD
                )
            elif cfg.CONNECTION_TYPE == 'IP':
                logger.info("Using IP Connection")

                self._connection = IPConnection(
                    self.on_connection_message, host=cfg.IP_CONNECTION_HOST,
                    port=cfg.IP_CONNECTION_PORT, password=cfg.IP_CONNECTION_PASSWORD
                )
            else:
                raise AssertionError("Invalid connection type: {}".format(cfg.CONNECTION_TYPE))

            self._register_connection_handlers()

        return self._connection

    def _register_connection_handlers(self):
        self.connection.register_handler(EventMessageHandler(self.handle_event_message))
        self.connection.register_handler(ErrorMessageHandler(self.handle_error_message))

    async def connect(self):
        if self._connection:
            self.disconnect()  # socket needs to be also closed
        self.panel = None

        self.run_state = RunState.INIT
        logger.info("Connecting to interface")
        if not await self.connection.connect():
            self.run_state = RunState.ERROR
            logger.error('Failed to connect to interface')
            return False

        logger.info("Connecting to panel")

        if not self.panel:
            self.panel = create_panel(self)
            self.connection.variable_message_length(self.panel.variable_message_length)

        try:
            logger.info("Initiating communication")

            initiate_reply = await self.send_wait(
                self.panel.get_message('InitiateCommunication'), None, reply_expected=0x07
            )

            if initiate_reply:
                model = initiate_reply.fields.value.label.strip(b'\0 ').decode(cfg.LABEL_ENCODING)
                firmware_version = "{}.{} build {}".format(
                    initiate_reply.fields.value.application.version,
                    initiate_reply.fields.value.application.revision,
                    initiate_reply.fields.value.application.build
                )
                serial_number = hexlify(initiate_reply.fields.value.serial_number).decode()

                logger.info("Found Panel {} version {}".format(model, firmware_version))
            else:
                raise ConnectionError("Panel did not replied to InitiateCommunication")

            logger.info("Starting communication")
            reply = await self.send_wait(self.panel.get_message('StartCommunication'),
                                         args=dict(source_id=0x02), reply_expected=0x00)

            if reply is None:
                raise ConnectionError("Panel did not replied to StartCommunication")

            if reply.fields.value.product_id is not None:
                self.panel = create_panel(self, reply.fields.value.product_id)  # Now we know what panel it is. Let's
                # recreate panel object.
                ps.sendMessage(
                    'panel_detected', panel=dict(
                        product_id=reply.fields.value.product_id, model=model,
                        firmware_version=firmware_version, serial_number=serial_number
                    )
                )

            result = await self.panel.initialize_communication(reply, cfg.PASSWORD)
            if not result:
                raise ConnectionError("Failed to initialize communication")

            if cfg.SYNC_TIME:
                await self.sync_time()

            if cfg.DEVELOPMENT_DUMP_MEMORY:
                if hasattr(self.panel, 'dump_memory') and callable(self.panel.dump_memory):
                    logger.warning("Requested memory dump. Dumping...")

                    await self.panel.dump_memory()
                    logger.warning("Memory dump completed. Exiting pai.")
                    raise SystemExit()
                else:
                    logger.warning("Requested memory dump, but current panel type does not support it yet.")

            logger.info("Loading definitions")
            definitions = await self.panel.load_definitions()
            ps.sendMessage('definitions_loaded', data=definitions)

            logger.info("Loading labels")
            labels = await self.panel.load_labels()
            ps.sendMessage('labels_loaded', data=labels)

            logger.info("Connection OK")
            self.run_state = RunState.RUN
            self.request_status_refresh()  # Trigger status update

            ps.sendMessage('connected')
            return True
        except asyncio.TimeoutError as e:
            logger.error("Timeout while connecting to panel: %s" % str(e))
        except ConnectionError as e:
            logger.error("Failed to connect: %s" % str(e))

        self.run_state = RunState.ERROR

        return False

    async def sync_time(self):
        logger.debug("Synchronizing panel time")

        now = time.localtime()
        args = dict(century=int(now.tm_year / 100), year=int(now.tm_year % 100),
                    month=now.tm_mon, day=now.tm_mday, hour=now.tm_hour, minute=now.tm_min)

        reply = await self.send_wait(self.panel.get_message('SetTimeDate'), args, reply_expected=0x03, timeout=10)
        if reply is None:
            logger.warning("Could not set panel time")
        else:
            logger.info("Panel time synchronized")

    async def loop(self):
        logger.debug("Loop start")
        
        replies_missing = 0

        while self.run_state not in (RunState.STOP, RunState.ERROR):
            tstart = time.time()
            if self.run_state == RunState.RUN:
                try:
                    await self.busy.acquire()
                    result = await asyncio.gather(*self.panel.get_status_requests())
                    merged = deep_merge(*result, extend_lists=True, initializer={})
                    self.work_loop.call_soon(self._process_status, merged)
                    replies_missing = max(0, replies_missing - 1)
                except ConnectionError:
                    raise
                except StatusRequestException:
                    replies_missing += 1
                    if replies_missing > 3:
                        logger.error("Lost communication with panel")
                        self.disconnect()
                except Exception:
                    logger.exception("Loop")
                finally:
                    self.busy.release()

                if replies_missing > 0:
                    logger.debug("Loop: Replies missing: {}".format(replies_missing))

            # cfg.Listen for events

            max_wait_time = max((tstart + cfg.KEEP_ALIVE_INTERVAL) - time.time(), 0)
            try:
                await asyncio.wait_for(self.loop_wait_event.wait(), max_wait_time)
            except asyncio.TimeoutError:
                # It is fine to timeout to go to the next loop
                pass
            finally:
                self.loop_wait_event.clear()

    @staticmethod
    def _process_status(raw_status: Container) -> None:
        status = convert_raw_status(raw_status)

        for limit_key, limit_arr in cfg.LIMITS.items():
            if limit_key not in status:
                continue

            status[limit_key].filter(limit_arr)

        #     # TODO: throttle power update messages
        #     if time.time() - self.last_power_update >= cfg.POWER_UPDATE_INTERVAL:
        #         force = PublishPropertyChange.YES if cfg.PUSH_POWER_UPDATE_WITHOUT_CHANGE else PublishPropertyChange.NO

        if cfg.LOGGING_DUMP_STATUS:
            logger.debug("properties: %s", status)

        ps.sendMessage('status_update', status=status)

    def request_status_refresh(self):
        self.loop_wait_event.set()

    def on_connection_message(self, message: bytes):
        self.connection.schedule_raw_message_handling(message)

        if not self.panel:
            return
        try:
            recv_message = self.panel.parse_message(message, direction='frompanel')

            if cfg.LOGGING_DUMP_MESSAGES:
                logger.debug('Message received: %s', recv_message)

            # No message
            if recv_message is None:
                logger.debug("Unknown message: %s" % (" ".join("{:02x} ".format(c) for c in message)))
                return

            if self.run_state != RunState.PAUSE:
                self.connection.schedule_message_handling(recv_message)  # schedule handling in the loop
        except Exception as e:
            logger.exception("Error parsing message")

    async def send_wait(self,
                        message_type=None,
                        args=None,
                        message=None,
                        retries=5,
                        timeout=0.5,
                        reply_expected=None) -> Optional[Container]:

        # Connection closed
        if not self.connection.connected:
            raise ConnectionError('Not connected')

        if message is None and message_type is not None:
            message = message_type.build(dict(fields=dict(value=args)))

        retry = 0

        while retry <= retries:
            if retry > 0:
                logger.debug('Request retry (%d/%d)', retry, retries)
            retry += 1

            t1 = time.time()
            result = 'unknown'
            try:
                async with self.request_lock:
                    if message is not None:
                        self.connection.write(message)

                    if reply_expected is not None:
                        if isinstance(reply_expected, Callable):
                            reply = await self.connection.wait_for_message(reply_expected, timeout=timeout * 2)
                        elif isinstance(reply_expected, Iterable):
                            reply = await self.connection.wait_for_message(
                                lambda m: any(m.fields.value.po.command == expected for expected in reply_expected),
                                timeout=timeout*2
                            )
                        else:
                            reply = await self.connection.wait_for_message(
                                lambda m: m.fields.value.po.command == reply_expected, timeout=timeout * 2
                            )

                        result = 'ok'
                        if reply:
                            return reply
            except asyncio.TimeoutError:
                result = 'timeout'
            except Exception:
                result = 'exception'
                raise
            finally:
                logger.debug('send/receive %s in %.4f s', result, time.time() - t1)

        return None  # Probably it needs to throw an exception instead of returning None

    async def control_zone(self, zone: str, command: str) -> bool:
        command = command.lower()
        logger.debug("Control Zone: {} - {}".format(zone, command))

        zones_selected = self.storage.get_container('zone').select(zone)  # type: Sequence[int]

        # Not Found
        if len(zones_selected) == 0:
            logger.error('No zones selected')
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_zones(zones_selected, command)
        except NotImplementedError:
            logger.error('control_zones is not implemented for this alarm type')
        except asyncio.CancelledError:
            logger.error('control_zones canceled')
        except asyncio.TimeoutError:
            logger.error('control_zones timeout')

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    async def control_partition(self, partition: str, command: str) -> bool:
        command = command.lower()
        logger.debug("Control Partition: {} - {}".format(partition, command))

        partitions_selected = self.storage.get_container('partition').select(partition)  # type: Sequence[int]

        # Not Found
        if len(partitions_selected) == 0:
            logger.error('No partitions selected')
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_partitions(partitions_selected, command)
        except NotImplementedError:
            logger.error('control_partitions is not implemented for this alarm type')
        except asyncio.CancelledError:
            logger.error('control_partitions canceled')
        except asyncio.TimeoutError:
            logger.error('control_partitions timeout')

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    async def control_output(self, output, command) -> bool:
        command = command.lower()
        logger.debug("Control Output: {} - {}".format(output, command))

        outputs_selected = self.storage.get_container('pgm').select(output)

        # Not Found
        if len(outputs_selected) == 0:
            logger.error('No outputs selected')
            return False

        # Apply state changes
        accepted = False
        try:
            accepted = await self.panel.control_outputs(outputs_selected, command)
        except NotImplementedError:
            logger.error('control_outputs is not implemented for this alarm type')
        except asyncio.CancelledError:
            logger.error('control_outputs canceled')
        except asyncio.TimeoutError:
            logger.error('control_outputs timeout')
        # Apply state changes

        # Refresh status
        self.request_status_refresh()  # Trigger status update

        return accepted

    def get_label(self, label_type: str, label_id) -> Optional[str]:
        el = self.storage.get_container_object(label_type, label_id)
        if el:
            return el.get("label")

    def handle_event_message(self, message: Container = None):
        """Process cfg.Live Event Message and dispatch it to the interface module"""
        try:
            try:
                evt = LiveEvent(event=message, event_map=self.panel.event_map, label_provider=self.get_label)
            except AssertionError as e:
                logger.debug("Error creating event")
                return

            element = self.storage.get_container_object(evt.type, evt.id)

            # Temporary to catch labels/properties in wrong places
            # TODO: REMOVE
            if message is not None:
                if not evt.id:
                    logger.debug(
                        "Missing element ID in {}/{}, m/m: {}/{}, message: {}".format(
                            evt.type, evt.label or '?', evt.major, evt.minor, evt.message
                        )
                    )
                else:
                    if not element:
                        logger.warning("Missing element with ID {} in {}/{}".format(evt.id, evt.type, evt.label))
                    else:
                        for k in evt.change:
                            if k not in element:
                                logger.warning("Missing property {} in {}/{}".format(k, evt.type, evt.label))
                        if evt.label != element.get("label"):
                            logger.warning(
                                "Labels differ {} != {} in {}/{}".format(
                                    element.get("label"), evt.label, evt.type, evt.label
                                )
                            )
            # Temporary end
            
            # The event has changes. Update the state
            if len(evt.change) > 0 and element:
                self.storage.update_container_object(evt.type, evt.id, evt.change)

            ps.sendEvent(evt)

        except Exception as e:
            logger.exception("Handle live event")

    def handle_error_message(self, message):
        """Handle ErrorMessage"""
        error_enum = message.fields.value.message

        if error_enum == 'panel_not_connected':
            self.disconnect()
        else:
            message = self.panel.get_error_message(error_enum)
            logger.error("Got ERROR Message: {}".format(message))

    def disconnect(self):
        logger.info("Disconnecting from the Alarm Panel")
        self.run_state = RunState.STOP

        self._clean_session()
        if self.connection.connected:
            self.connection.close()
            logger.info("Disconnected from the Alarm Panel")

    async def pause(self):
        logger.info("Pausing PAI")
        if self.run_state == RunState.RUN:
            logger.info("Pausing from the Alarm Panel")
            self.run_state = RunState.PAUSE
            # EVO IP150 IP Interface does not work if we send this
            # await self.send_wait(self.panel.get_message('CloseConnection'), None)

    async def resume(self):
        logger.info("Resuming PAI")
        if self.run_state == RunState.PAUSE:
            await self.connect()

    def _clean_session(self):
        logger.info("Clean Session")
        if self.connection.connected:
            if not self.panel:
                panel = create_panel(self)
            else:
                panel = self.panel

            logger.info("Cleaning previous session. Closing connection")
            # Write directly as this can be called from other contexts

            self.connection.write(panel.get_message('CloseConnection').build(dict()))

    def _on_labels_load(self, data):
        for k, d in data.items():
            self.storage.get_container(k).deep_merge(d)

    def _on_definitions_load(self, data):
        for k, d in data.items():
            self.storage.get_container(k).deep_merge(d)

    def _on_status_update(self, status):
        """
        Calls update_properties
        :param status:
        :return:
        """
        if 'troubles' in status:
            self._process_trouble_statuses(status['troubles'])

        for element_type, element_items in status.items():
            if element_type in ['troubles']:  # troubles was already parsed
                continue
            for element_item_key, element_item_status in element_items.items():
                if isinstance(element_item_status, (dict, list,)):
                    self.storage.update_container_object(element_type, element_item_key, element_item_status)
                else:
                    logger.debug("%s/%s:%s ignored", element_type, element_item_key, element_item_status)

        self._update_partition_states()

    def _process_trouble_statuses(self, trouble_statuses):
        global_trouble = False
        for t_key, t_status in trouble_statuses.items():
            if not isinstance(t_status, bool):
                logger.error("Trouble %s has not boolean state: %s", t_key, t_status)
                continue

            self.storage.update_container_object('system', 'troubles', {t_key: t_status})

            global_trouble = global_trouble or t_status

        self.storage.update_container_object('system', 'troubles', {'trouble': global_trouble})

    def _update_partition_states(self):
        """
        current_state is fully HomeAssistant compatible. Check HASS manual before making any changes.
        """
        for _, properties in self.storage.get_container('partition').items():
            change = {}
            if any([
                properties.get('fire_alarm'),
                properties.get('audible_alarm'),
                properties.get('silent_alarm'),
                properties.get('panic_alarm')
            ]):
                change["current_state"] = 'triggered'
            elif properties.get('arm'):
                if properties.get('arm_stay'):
                    change["current_state"] = 'armed_home'
                elif properties.get('arm_sleep'):
                    change["current_state"] = 'armed_night'
                elif properties.get('arm_away'):
                    change["current_state"] = 'armed_away'
                else:
                    change["current_state"] = 'armed_away'

                change["target_state"] = change["current_state"]
            else:
                change["target_state"] = change["current_state"] = 'disarmed'

            if properties.get('exit_delay'):  # Redefine if pending
                change["current_state"] = 'pending'

            self.storage.update_container_object('partition', properties['key'], change)

    def _on_event(self, event: Event):
        if cfg.LOGGING_DUMP_EVENTS:
            logger.debug("LiveEvent: {}".format(event))

        event.call_hook(storage=self.storage, alarm=self)

        if isinstance(event, LiveEvent):
            self._update_partition_states()

    def _on_property_change(self, change: Change):
        if change.initial:
            return

        try:
            event = ChangeEvent(
                change_object=change, property_map=self.panel.property_map, label_provider=self.get_label
            )
            if cfg.LOGGING_DUMP_EVENTS:
                logger.debug("ChangeEvent: {}".format(event))
            ps.sendEvent(event)
        except AssertionError:
            logger.debug("Could not create event from change")



#!/usr/bin/env python3
import argparse
import binascii
import yaml
from collections import OrderedDict

from paradox.lib.crypto import decrypt


def ordered_load(stream, Loader=yaml.loader.SafeLoader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.load(stream, OrderedLoader)


def decrypt_file(file, password):
    PASSWORD2 = None
    try:
        data = ordered_load(file, yaml.loader.SafeLoader)

        n = 0
        for key, value in data.items():
            header = value[0:16]
            body = value[16:]
            decrypted = decrypt(body, password if n < 2 else PASSWORD2)
            if n < 2 and "peer1_" in key:
                PASSWORD2 = decrypted[1:17]
                print(len(PASSWORD2))

            print("PC->IP: " if "peer0_" in key else "IP->PC: ", binascii.hexlify(header),
                  binascii.hexlify(decrypted), decrypted)

            if "peer1_" in key:
                print('----end %s-------------------------------------------------------------' % key)
            n += 1

    except yaml.YAMLError as exc:
        print(exc)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=argparse.FileType('r'),
                        help="YAML file to parse. In wireshark right click on the first "
                             "package, 'Follow->TCP Stream', 'Show and save data as': "
                             "'YAML', copy contents to a file.")
    parser.add_argument("password", type=str, help="IP Module password for decryption")

    args = parser.parse_args()

    decrypt_file(args.file, args.password.encode('utf8'))


if __name__ == '__main__':
    main()



# coding: utf-8

# Adapted from code in https://github.com/konomae/stunpy

import random
import socket
import binascii
import struct

STUN_PORT = 3478

FAMILY_IPv4 = b'\x01'
FAMILY_IPv6 = b'\x02'

# 0b00: Request
# 0b01: Binding
BINDING_REQUEST_SIGN = b'\x00\x01'  # 16bit (2bytes)

BINDING_RESPONSE_ERROR = b'\x01\x11'
BINDING_RESPONSE_SUCCESS = b'\x01\x01'
MAGIC_COOKIE = b'\x21\x12\xA4\x42'  # 32bit (4bytes)

CHANGE_RESPONSE_SUCCESS = b'\x01\x03'

CONNECT_REQUEST = b'\x00\x0a'
CONNECT_RESPONSE_SUCCESS = b'\x01\x0a'

CONNECTION_BIND_REQUEST = b'\x00\x0b'
CONNECTION_BIND_SUCCESS = b'\x01\x0b'

CONNECTION_REFRESH_REQUEST = b'\x00\x04'

# STUN Attribute Registry
MAPPED_ADDRESS = b'\x00\x01'
RESPONSE_ADDRESS = b'\x00\x02'
CHANGE_REQUEST = b'\x00\x03'
SOURCE_ADDRESS = b'\x00\x04'
CHANGED_ADDRESS = b'\x00\x05'
USERNAME = b'\x00\x06'
PASSWORD = b'\x00\x07'
MESSAGE_INTEGRITY = b'\x00\x08'
ERROR_CODE = b'\x00\x09'
UNKNOWN_ATTRIBUTES = b'\x00\x0A'
REFLECTED_FROM = b'\x00\x0B'
REALM = b'\x00\x14'
NONCE = b'\x00\x15'
XOR_MAPPED_ADDRESS = b'\x80\x82'
XOR_PEER_ADDRESS = b'\x00\x12'
REQUESTED_TRANSPORT = b'\x00\x19'
TRANSPORT_TCP = b'\x06'
LIFETIME = b'\x00\x0d'
REQUESTED_ADDRESS_TYPE = b'\x00\x17'
RESERVED3 = b'\x00\x00\x00'
RESERVED1 = b'\x00'
CONNECTION_ID = b'\x00\x2a'

STUN_ATTRIBUTE_NAMES = {
    MAPPED_ADDRESS: 'MAPPED-ADDRESS',
    RESPONSE_ADDRESS: 'RESPONSE-ADDRESS',
    CHANGE_REQUEST: 'CHANGE-REQUEST',
    SOURCE_ADDRESS: 'SOURCE-ADDRESS',
    CHANGED_ADDRESS: 'CHANGED-ADDRESS',
    USERNAME: 'USERNAME',
    PASSWORD: 'PASSWORD',
    MESSAGE_INTEGRITY: 'MESSAGE-INTEGRITY',
    ERROR_CODE: 'ERROR-CODE',
    UNKNOWN_ATTRIBUTES: 'UNKNOWN-ATTRIBUTES',
    REFLECTED_FROM: 'REFLECTED-FROM',
    REALM: 'REALM',
    NONCE: 'NONCE',
    XOR_MAPPED_ADDRESS: 'XOR-MAPPED-ADDRESS',
    b'\x00\x20': 'XOR-MAPPED-ADDRESS',
}


def generate_transaction_id():
    tid = []
    for i in range(24):  # 96bits (12bytes)
        tid.append(random.choice('0123456789ABCDEF'))
    return binascii.a2b_hex(''.join(tid))


def build_binding_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception('Invalid transaction id')

    body_length = b'\x00\x00'
    return b''.join([BINDING_REQUEST_SIGN, body_length, MAGIC_COOKIE, transaction_id])


def build_change_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception('Invalid transaction id')

    body_length = b'\x00\x18'
    return b''.join([CHANGE_REQUEST, body_length, MAGIC_COOKIE, transaction_id,
                     REQUESTED_TRANSPORT, b'\x00\x04', TRANSPORT_TCP, RESERVED3,
                     LIFETIME, b'\x00\x04', b'\x00\x00\x02\x58',
                     REQUESTED_ADDRESS_TYPE, b'\x00\x04', FAMILY_IPv4, RESERVED3])


def build_connection_bind_request(transaction_id, connection_id):
    if len(transaction_id) != 12:
        raise Exception('Invalid transaction id')

    if len(connection_id) != 4:
        raise Exception('Invalid connection id {}'.format(len(connection_id)))

    body_length = b'\x00\x08'
    return b''.join([CONNECTION_BIND_REQUEST, body_length, MAGIC_COOKIE, transaction_id,
                     CONNECTION_ID, b'\x00\x04', connection_id])


def build_connection_refresh_request(transaction_id):
    if len(transaction_id) != 12:
        raise Exception('Invalid transaction id')

    body_length = b'\x00\x08'
    return b''.join([CONNECTION_REFRESH_REQUEST, body_length, MAGIC_COOKIE, transaction_id,
                     LIFETIME, b'\x00\x04', b'\x00\x00\x02\x58'])


def build_connect_request(transaction_id, xoraddr=None):
    if len(transaction_id) != 12:
        raise Exception('Invalid transaction id')

    if xoraddr is None:
        raise Exception("Invalid connect address")

    body_length = b'\x00\x0c'
    return b''.join([CONNECT_REQUEST, body_length, MAGIC_COOKIE, transaction_id,
                     XOR_PEER_ADDRESS, b'\x00\x08', xoraddr])


def validate_response(buf, transaction_id):
    if not buf or len(buf) < 20:
        raise Exception('Response too short')

    response_magic_cookie = buf[4:8]
    if MAGIC_COOKIE != response_magic_cookie:
        raise Exception('Invalid magic cookie')

    response_transaction_id = buf[8:20]
    if transaction_id != response_transaction_id:
        raise Exception('invalid transaction id')


def ip_to_bytes(ip, xor):
    octets = [binascii.a2b_hex('%02x' % int(o)) for o in ip.split('.')]
    addr_int = struct.unpack('!I', b''.join(octets))[0]

    if xor:
        magicCookieBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE), 16)
        addr_int = magicCookieBytesInt ^ addr_int

    addr_bytes = binascii.a2b_hex('%08x' % addr_int)
    return addr_bytes


def port_to_bytes(port, xor):
    if xor:
        magicCookieHighBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE[:2]), 16)
        port = magicCookieHighBytesInt ^ port

    port_bytes = binascii.a2b_hex('%04x' % port)
    return port_bytes


def read_mapped_address(attr_type, attr_body, attr_len):
    assert attr_type in (MAPPED_ADDRESS, XOR_MAPPED_ADDRESS, b'\x00\x20')
    assert attr_body[:1] == b'\x00'

    family_bytes = attr_body[1:2]
    port_bytes = attr_body[2:4]
    addr_bytes = attr_body[4:attr_len]
    xor = attr_type in (XOR_MAPPED_ADDRESS, b'\x00\x20')

    family_text = ''
    assert family_bytes in (FAMILY_IPv4, FAMILY_IPv6)
    if family_bytes == FAMILY_IPv4:
        family_text = 'IPv4'
    elif family_bytes == FAMILY_IPv6:
        family_text = 'IPv6'

    # TODO: IPv6s
    port = int(binascii.b2a_hex(port_bytes), 16)
    addr_int = int(binascii.b2a_hex(addr_bytes), 16)

    if xor:
        # port
        magicCookieHighBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE[:2]), 16)
        port = magicCookieHighBytesInt ^ port

        # addr
        magicCookieBytesInt = int(binascii.b2a_hex(MAGIC_COOKIE), 16)
        addr_int = magicCookieBytesInt ^ addr_int

    octets = struct.pack('!I', addr_int)
    ip = '.'.join([str(c) for c in octets])
    return dict(name=STUN_ATTRIBUTE_NAMES[attr_type], ip=ip, port=port, family=family_text)


def read_attributes(attributes, body_length):
    pos = 0
    parsed_attributes = []
    while pos < body_length:
        attr_type = attributes[pos:pos + 2]  # 16bit (2bytes)
        attr_len = int(binascii.b2a_hex(attributes[pos + 2:pos + 4]), 16)  # 16bit (2bytes)
        attr_body = attributes[pos + 4:pos + 4 + attr_len]

        if attr_type in (MAPPED_ADDRESS, XOR_MAPPED_ADDRESS, b'\x00\x20'):
            parsed_attributes.append(read_mapped_address(attr_type, attr_body, attr_len))
        else:
            if attr_type == b'\x80\x22':
                attr_body = attr_body.decode('utf-8')
            else:
                attr_body = attr_body.hex()
            parsed_attributes.append(dict(
                name=STUN_ATTRIBUTE_NAMES.get(attr_type),
                attr_type=attr_type.hex(),
                attr_body=attr_body,
                attr_len=attr_len
            ))

        attr_head = (2 + 2)  # attr_type + attr_len
        remain = attr_len % 4
        padding = 4 - remain if remain else 0
        pos += attr_head + attr_len + padding

    return parsed_attributes


def is_error(body):
    for element in body:
        if element['name'] == 'ERROR-CODE':
            return True

    return False


def get_error(body):
    for element in body:
        if element['name'] == 'ERROR-CODE':
            body = binascii.unhexlify(element['attr_body'])
            return {'class': body[:3], 'code': body[3], 'reason': body[4:]}

    return False


class StunClient(object):
    def __init__(self, host, port=STUN_PORT):
        self.sock = None
        self.transaction_id = None
        self.req = None
        self.transaction_id = generate_transaction_id()

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind(('0.0.0.0', 0))
        self.host = host
        self.port = port
        self.sock.connect((host, port))

    def send_tcp_change_request(self):
        self.req = build_change_request(self.transaction_id)
        self.sock.send(self.req)

    def send_binding_request(self):
        self.req = build_binding_request(self.transaction_id)
        self.sock.send(self.req)

    def send_connect_request(self, host=None, port=None, xoraddr=None):
        if host is not None and port is not None:
            xorhost = ip_to_bytes(host, True)
            xorport = port_to_bytes(port, True)
            addr = b''.join([xorport, xorhost])
        elif xoraddr is not None:
            addr = xoraddr
        else:
            raise Exception("Invalid arguments")

        self.req = build_connect_request(self.transaction_id, addr)
        self.sock.send(self.req)

    def send_connection_bind_request(self, connection_id):
        self.req = build_connection_bind_request(self.transaction_id, connection_id)
        self.sock.send(self.req)

    def receive_response(self):
        buf = self.sock.recv(2048)
        validate_response(buf, self.transaction_id)

        body_length = int(binascii.b2a_hex(buf[2:4]), 16)
        attributes = buf[20:]
        assert len(attributes) == body_length

        return read_attributes(attributes, body_length)

    def send_refresh_request(self):
        self.req = build_connection_refresh_request(self.transaction_id)
        self.sock.send(self.req)

    def get_socket(self):
        return self.sock

    def close(self):
        if self.sock:
            self.sock.close()



import logging
logger = logging.getLogger('PAI').getChild(__name__)

MODULES = {
    'construct': dict(mandatory=True, desc='basic operation', install_name='construct>=2.9.43'),
    'argparse': dict(mandatory=True, desc='basic operation', install_name='argparse>=1.4.0'),
    'pyserial': dict(mandatory=False, desc='the serial connection', install_name='pyserial>=3.4'),
    'serial_asyncio': dict(mandatory=False, desc='the serial connection', install_name='pyserial-asyncio>=0.4'),
    'pushbullet': dict(mandatory=False, desc='the Pushbullet interface', install_name='pushbullet.py=>0.11.0'),
    'requests': dict(mandatory=False, desc='the IP150 connection', install_name='requests>=2.20.0'),
    'ws4py': dict(mandatory=False, desc='the Pushbullet interface', install_name='ws4py>=0.4.2'),
    'yaml': dict(mandatory=False, desc='the IP150 connection', install_name='yaml'),
    'chump': dict(mandatory=False, desc='the Pushover interface', install_name='chump>=1.6.0'),
    'pydbus': dict(mandatory=False, desc='the Signal interface', install_name='pydbus>=0.6.0'),
    'gi': dict(mandatory=False, desc='the Signal interface', install_name='gi>=1.2'),
}


def import_error_help(error):
    logger.error("Could not import Python3 module '{}': {}\n".format(error.name, error))

    if error.name in MODULES:
        m = MODULES[error.name]
        logger.error(
            "The module is required for {} and IS{} mandatory.".format(m['desc'], ' NOT' if not m['mandatory'] else '')
        )
        if not m['mandatory']:
            logger.error('If you do not require such functionality, you can disable it in the config file,')
            logger.error('and skip installing the module.\n')
        logger.error("To install ONLY this module, execute: \n")
        logger.error(" pip3 install '{}'\n".format(m['install_name']))
    
    logger.error("To install ALL modules required, go to the main project folder and execute:\n")
    logger.error(" pip3 -r requirements.txt\n")
    logger.error("Take in consideration that the 'requirements.txt' file only has the most common modules enabled.")
    logger.error("Uncomment the entries as required for your setup.")
    logger.error("ATTENTION: If this module is not listed there, please report the bug.")
    import sys
    
    sys.exit(-1)




from construct import Struct, BitStruct, Const, Nibble, Padding, Bytes, RawCopy, Int8ub, \
    Default, Enum, Flag, BitsInteger, Int16ub

from .common import ProductIdEnum, CommunicationSourceIDEnum, HexInt, PacketLength, PacketChecksum

InitiateCommunication = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(7, Nibble),
                "reserved0" / Const(2, Nibble)
            ),
            "reserved1" / Padding(35)
        )
    ),
    "checksum" / PacketChecksum(Bytes(1))
)

InitiateCommunicationResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct(
                "command" / Const(7, Nibble),
                "message_center" / Nibble
            ),
            "new_protocol" / Const(0xFF, Int8ub),
            "protocol_id" / Int8ub,
            "protocol" / Struct(
                "version" / Int8ub,
                "revision" / Int8ub,
                "build" / Int8ub
            ),
            "family_id" / Int8ub,
            "product_id" / ProductIdEnum,
            "talker" / Enum(Int8ub,
                            BOOT_LOADER=0,
                            CONTROLLER_APPLICATION=1,
                            MODULE_APPLICATION=2),
            "application" / Struct(
                "version" / HexInt,
                "revision" / HexInt,
                "build" / HexInt),
            "serial_number" / Bytes(4),
            "hardware" / Struct(
                "version" / Int8ub,
                "revision" / Int8ub),
            "bootloader" / Struct(
                "version" / Int8ub,
                "revision" / Int8ub,
                "build" / Int8ub,
                "day" / Int8ub,
                "month" / Int8ub,
                "year" / Int8ub),
            "processor_id" / Int8ub,
            "encryption_id" / Int8ub,
            "reserved0" / Bytes(2),
            "label" / Bytes(8)
        )
    ),
    "checksum" / PacketChecksum(Bytes(1)))

StartCommunication = Struct("fields" / RawCopy(
    Struct(
        "po" / Struct("command" / Const(0x5F, Int8ub)),
        "validation" / Const(0x20, Int8ub),
        "_not_used0" / Padding(31),
        "source_id" / Default(CommunicationSourceIDEnum, 1),
        "user_id" / Struct(
            "high" / Default(Int8ub, 0),
            "low" / Default(Int8ub, 0)),
    )), "checksum" / PacketChecksum(Bytes(1)))

StartCommunicationResponse = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / BitStruct("command" / Const(0, Nibble),
                             "status" / Struct(
                                 "reserved" / Flag,
                                 "alarm_reporting_pending" / Flag,
                                 "Winload_connected" / Flag,
                                 "NeWare_connected" / Flag)
                             ),
            "_not_used0" / Bytes(3),
            "product_id" / ProductIdEnum,
            "firmware" / Struct(
                "version" / Int8ub,
                "revision" / Int8ub,
                "build" / Int8ub),
            "panel_id" / Int16ub,
            "_not_used1" / Bytes(5),
            "transceiver" / Struct(
                "firmware_build" / Int8ub,
                "family" / Int8ub,
                "firmware_version" / Int8ub,
                "firmware_revision" / Int8ub,
                "noise_floor_level" / Int8ub,
                "status" / BitStruct(
                    "_not_used" / BitsInteger(6),
                    "noise_floor_high" / Flag,
                    "constant_carrier" / Flag,
                ),
                "hardware_revision" / Int8ub,
            ),
            "_not_used2" / Bytes(14),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1))
)

CloseConnection = Struct(
    "fields" / RawCopy(
        Struct(
            "po" / Struct(
                "command" / Const(0x70, Int8ub)
            ),
            "length" / PacketLength(Int8ub),
            "_not_used0" / Padding(34),
        )
    ),
    "checksum" / PacketChecksum(Bytes(1))
)
